import os
from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu

from .hmc import init_bold_hmc_wf,EstimateMotionParams
from .bold_ref import init_bold_reference_wf
from .resampling import init_bold_preproc_trans_wf,init_mask_preproc_trans_wf
from .stc import init_bold_stc_wf
from .inho_correction import init_inho_correction_wf
from .registration import init_cross_modal_reg_wf
from nipype.interfaces.utility import Function
from .utils import apply_despike, log_transform_nii
from ..utils import ResampleMask

def init_bold_main_wf(opts, output_folder, number_functional_scans, inho_cor_only=False, name='bold_main_wf'):
    """
    This workflow controls the functional preprocessing stages of the pipeline when both
    functional and anatomical images are provided.

    **Parameters**

        opts
            parser options for preprocess
        inho_cor_only
            whether to run the bias correction steps, or further processing steps.

    **Inputs**

        bold
            Input BOLD series NIfTI file
        coreg_anat
            Anatomical reference for BOLD alignment
        coreg_mask
            Brain mask for anatomical reference
        WM_mask
            WM mask inherited from the common space registration
        CSF_mask
            CSF mask inherited from the common space registration
        vascular_mask
            vascular mask inherited from the common space registration
        unbiased_to_atlas_affine
            affine transform from the dataset template space to the commonspace space
        unbiased_to_atlas_warp
            non-linear transform from the dataset template space to the commonspace space
        anat_to_unbiased_affine
            affine transform from the subject anatomical to the dataset template space
        anat_to_unbiased_warp
            non-linear transform from the subject anatomical to the dataset template space
        commonspace_ref
            commonspace anatomical template

    **Outputs**

        input_bold
            The provided input BOLD file
        bold_ref
            Initial EPI median volume subsequently used as 3D reference EPI volume
        motcorr_params
            motion parameters file provided from antsMotionCorr
        init_denoise
            Corrected 3D ref EPI after initial correction step
        denoise_mask
            resampled mask used for final denoising
        corrected_EPI
            3D reference EPI volume after bias field correction
        output_warped_bold
            Bias field corrected 3D EPI volume warped to the anatomical space
        bold_to_anat_affine
            affine transform from the EPI space to the anatomical space
        bold_to_anat_warp
            non-linear transform from the EPI space to the anatomical space
        bold_to_anat_inverse_warp
            inverse non-linear transform from the EPI space to the anatomical space
        resampled_bold
            Original BOLD timeseries resampled through motion realignment and
            susceptibility distortion correction based on registration to the
            anatomical image
        resampled_ref_bold
            3D median EPI volume from the resampled native BOLD timeseries
        motion_params_csv
            .csv file with measured motion timecourses, used as regressors for confound
            correction: 6 rigid body motion parameters + their first temporal derivate 
            + the 12 parameters squared (24 motion parameters)
        FD_voxelwise
            Voxelwise framewise displacement (FD) measures that can be integrated
            to future confound regression.
            These measures are computed from antsMotionCorrStats.
        pos_voxelwise
            Voxel distancing across time based on rigid body movement parameters,
            which can be integrated for a voxelwise motion regression
            These measures are computed from antsMotionCorrStats.
        FD_csv
            .csv file with global framewise displacement (FD) measures
        EPI_brain_mask
            EPI brain mask for resampled bold
        EPI_WM_mask
            EPI WM mask for resampled bold
        EPI_CSF_mask
            EPI CSF mask for resampled bold
        commonspace_bold
            Motion and SDC-corrected EPI timeseries resampled into common space
            by applying transforms from the anatomical common space registration
        commonspace_mask
            EPI brain mask for commonspace bold
        commonspace_WM_mask
            EPI WM mask for commonspace bold
        commonspace_CSF_mask
            EPI CSF mask for commonspace bold
        commonspace_vascular_mask
            EPI vascular mask for commonspace bold
    """

    workflow = pe.Workflow(name=name)

    inputnode = pe.Node(niu.IdentityInterface(
                fields=['bold', 'inho_cor_anat', 'inho_cor_mask', 'coreg_anat', 'coreg_mask',
                        'anat_to_commonspace_transform_list','anat_to_commonspace_inverse_list',
                        'commonspace_to_anat_transform_list','commonspace_to_anat_inverse_list',
                        'commonspace_ref']),
                        name="inputnode")

    outputnode = pe.Node(niu.IdentityInterface(
                fields=['input_bold', 'bold_ref', 'motcorr_params', 'init_denoise', 'denoise_mask', 'corrected_EPI',
                        'output_warped_bold', 'bold_to_anat_affine', 'bold_to_anat_warp', 'bold_to_anat_inverse_warp',
                        'native_bold', 'native_bold_ref', 'native_brain_mask', 'native_WM_mask', 'native_CSF_mask', 'native_vascular_mask',
                        'motion_params_csv', 'FD_voxelwise', 'pos_voxelwise', 'FD_csv', 'commonspace_bold', 'commonspace_mask',
                        'commonspace_WM_mask', 'commonspace_CSF_mask', 'commonspace_vascular_mask',
                        'boldspace_brain_mask',
                        'hmc_qc_figure','hmc_qc_csv','hmc_qc_video',
                        ]),
                name='outputnode')

    boldbuffer = pe.Node(niu.IdentityInterface(fields=['bold_file']),
                         name="boldbuffer")

    # this node will serve as a relay of outputs from the inho_cor main_wf to the inputs for the rest of the main_wf for bold_only
    transitionnode = pe.Node(niu.IdentityInterface(fields=['bold_file', 'isotropic_bold_file', 'bold_ref', 'init_denoise', 'denoise_mask', 'corrected_EPI', 'log_bold']),
                             name="transitionnode")        

    if inho_cor_only or (not opts.bold_only):
        template_inputnode = pe.Node(niu.IdentityInterface(fields=['template_anat', 'template_mask']),
                                            name="template_inputnode")


        bold_reference_wf = init_bold_reference_wf(opts=opts)

        # calculate the number of threads used for modelbuild
        # the number cannot exceed --local_threads, but should use number_functional_scans*num_ITK_threads at least
        nthreads_modelbuild_bold = min(opts.local_threads, number_functional_scans*int(os.environ['RABIES_ITK_NUM_THREADS']))
        inho_cor_wf = init_inho_correction_wf(opts=opts, image_type='EPI', output_folder=output_folder, nthreads_modelbuild=nthreads_modelbuild_bold, name="bold_inho_cor_wf")


        if opts.log_transform:
            log_bold_node = pe.Node(Function(input_names=['in_nii'],
                                                            output_names=[
                                                                'log_nii'],
                                                            function=log_transform_nii),
                                                    name='log_bold_node')

            workflow.connect([
                (boldbuffer, log_bold_node, [
                    ('bold_file', 'in_nii'),
                    ]),
                (log_bold_node, transitionnode, [
                    ('log_nii', 'log_bold'),
                    ]),
                (log_bold_node, bold_reference_wf, [
                    ('log_nii', 'inputnode.bold_file'),
                    ]),
                ])
        else:
            workflow.connect([
                (boldbuffer, bold_reference_wf, [
                    ('bold_file', 'inputnode.bold_file'),
                    ]),
                ])

        if opts.apply_despiking:
            despike = pe.Node(Function(input_names=['in_file'],
                                                        output_names=['out_file'],
                                                        function=apply_despike),
                                            name='despike')
            workflow.connect([
                (inputnode, despike, [('bold', 'in_file')]),
                (despike, boldbuffer, [('out_file', 'bold_file')]),
                ])
        else:
            workflow.connect([
                (inputnode, boldbuffer, [('bold', 'bold_file')]),
                ])

        if opts.detect_dummy:
            def remove_dummy(bold_file):
                import SimpleITK as sitk
                import pathlib
                import os
                from rabies.utils import copyInfo_4DImage
                from rabies.preprocess_pkg.bold_ref import _get_vols_to_discard

                in_nii = sitk.ReadImage(bold_file)
                data_array = sitk.GetArrayFromImage(in_nii)
                n_volumes_to_discard = _get_vols_to_discard(in_nii)
                if (not n_volumes_to_discard == 0):
                    filename_split = pathlib.Path(bold_file).name.rsplit(".nii")
                    out_bold_file = os.path.abspath(
                        f'{filename_split[0]}_cropped_dummy.nii.gz')
                    img_array = data_array[n_volumes_to_discard:, :, :, :]

                    image_4d = copyInfo_4DImage(sitk.GetImageFromArray(
                        img_array, isVector=False), in_nii, in_nii)
                    sitk.WriteImage(image_4d, out_bold_file)
                else:
                    out_bold_file = bold_file
                return out_bold_file

            remove_dummy_node = pe.Node(Function(input_names=['bold_file'],
                                                            output_names=[
                                                                'out_bold_file'],
                                                            function=remove_dummy),
                                                    name='remove_dummy')
            workflow.connect([
                (boldbuffer, remove_dummy_node, [
                    ('bold_file', 'bold_file'),
                    ]),
                (remove_dummy_node, transitionnode, [
                    ('out_bold_file', 'bold_file'),
                    ]),
                ])
        else:
            workflow.connect([
                (boldbuffer, transitionnode, [
                    ('bold_file', 'bold_file'),
                    ]),
                ])

        workflow.connect([
            (inputnode, inho_cor_wf, [
                ('inho_cor_anat', 'inputnode.anat_ref'),
                ('inho_cor_mask', 'inputnode.anat_mask'),
                ('bold', 'inputnode.name_source'),
                ]),
            (template_inputnode, inho_cor_wf, [
                ("template_anat", "template_inputnode.template_anat"),
                ("template_mask", "template_inputnode.template_mask"),
                ]),
            (bold_reference_wf, inho_cor_wf, [
                ('outputnode.ref_image', 'inputnode.target_img'),
                ]),
            (bold_reference_wf, transitionnode, [
                ('outputnode.ref_image', 'bold_ref'),
                ]),
            (inho_cor_wf, transitionnode, [
                ('outputnode.init_denoise', 'init_denoise'),
                ('outputnode.denoise_mask', 'denoise_mask'),
                ('outputnode.corrected', 'corrected_EPI'),
                ]),
            ])

    if inho_cor_only:
        return workflow

    bold_stc_wf = init_bold_stc_wf(opts=opts)

    # HMC on the BOLD
    bold_hmc_wf = init_bold_hmc_wf(opts=opts)

    estimate_motion_node = pe.Node(EstimateMotionParams(),
                                name='estimate_motion_node', mem_gb=2.3*opts.scale_min_memory)
    estimate_motion_node.plugin_args = {
        'qsub_args': f'-pe smp {str(2*opts.min_proc)}', 'overwrite': True}


    def prep_transforms_between_spaces(anat_to_commonspace_transform_list,anat_to_commonspace_inverse_list, bold_to_anat_warp, bold_to_anat_inverse_warp, bold_to_anat_affine, 
                                   commonspace_to_anat_transform_list, commonspace_to_anat_inverse_list, bold_only=False, bold_nativespace=False):
        if bold_only:
            # to generate commonspace timeseries
            bold_to_commonspace_transform_list = anat_to_commonspace_transform_list
            bold_to_commonspace_inverse_list = anat_to_commonspace_inverse_list
            # to generate nativespace timeseries
            bold_to_native_transform_list = []
            bold_to_native_inverse_list = []
            # to generate a mask in bold space
            commonspace_to_bold_transform_list = commonspace_to_anat_transform_list
            commonspace_to_bold_inverse_list = commonspace_to_anat_inverse_list
            # generic transform to move from native to common
            native_to_commonspace_transform_list = anat_to_commonspace_transform_list
            native_to_commonspace_inverse_list = anat_to_commonspace_inverse_list
            # generic transform to move from common to native
            commonspace_to_native_transform_list = commonspace_to_anat_transform_list
            commonspace_to_native_inverse_list = commonspace_to_anat_inverse_list
        else:
            # to generate commonspace timeseries
            bold_to_commonspace_transform_list = anat_to_commonspace_transform_list+[bold_to_anat_warp, bold_to_anat_affine]
            bold_to_commonspace_inverse_list = anat_to_commonspace_inverse_list+[0,0]
            # to generate a mask in bold space
            commonspace_to_bold_transform_list = [bold_to_anat_affine, bold_to_anat_inverse_warp]+commonspace_to_anat_transform_list
            commonspace_to_bold_inverse_list = [1,0]+commonspace_to_anat_inverse_list
            if bold_nativespace:
                # to generate nativespace timeseries
                bold_to_native_transform_list = []
                bold_to_native_inverse_list = []
                # generic transform to move from native to common
                native_to_commonspace_transform_list = bold_to_commonspace_transform_list
                native_to_commonspace_inverse_list = bold_to_commonspace_inverse_list
                # generic transform to move from common to native
                commonspace_to_native_transform_list = commonspace_to_bold_transform_list
                commonspace_to_native_inverse_list = commonspace_to_bold_inverse_list
            else:
                # to generate nativespace timeseries
                bold_to_native_transform_list = [bold_to_anat_warp, bold_to_anat_affine]
                bold_to_native_inverse_list = [0, 0]
                # generic transform to move from native to common
                native_to_commonspace_transform_list = anat_to_commonspace_transform_list
                native_to_commonspace_inverse_list = anat_to_commonspace_inverse_list
                # generic transform to move from common to native
                commonspace_to_native_transform_list = commonspace_to_anat_transform_list
                commonspace_to_native_inverse_list = commonspace_to_anat_inverse_list

        return [bold_to_commonspace_transform_list, bold_to_commonspace_inverse_list, 
                commonspace_to_bold_transform_list, commonspace_to_bold_inverse_list,
                native_to_commonspace_transform_list, native_to_commonspace_inverse_list, 
                commonspace_to_native_transform_list, commonspace_to_native_inverse_list,
                bold_to_native_transform_list, bold_to_native_inverse_list]

    output_transforms_between_spaces = ['bold_to_commonspace_transform_list', 'bold_to_commonspace_inverse_list', 
                                        'commonspace_to_bold_transform_list', 'commonspace_to_bold_inverse_list', 
                                        'native_to_commonspace_transform_list', 'native_to_commonspace_inverse_list', 
                                        'commonspace_to_native_transform_list', 'commonspace_to_native_inverse_list',
                                        'bold_to_native_transform_list', 'bold_to_native_inverse_list']
    prep_transforms_between_spaces_node = pe.Node(Function(input_names=['anat_to_commonspace_transform_list','anat_to_commonspace_inverse_list', 'bold_to_anat_warp', 'bold_to_anat_inverse_warp', 'bold_to_anat_affine', 
                                                                    'commonspace_to_anat_transform_list', 'commonspace_to_anat_inverse_list', 'bold_only', 'bold_nativespace'],
                                                    output_names=output_transforms_between_spaces,
                                                    function=prep_transforms_between_spaces),
                                            name='prep_transforms_between_spaces')
    prep_transforms_between_spaces_node.inputs.bold_only = opts.bold_only
    prep_transforms_between_spaces_node.inputs.bold_nativespace = opts.bold_nativespace

    # this node will hold onto the outputs of prep_transforms_between_spaces so that they are available in later RABIES stages
    def transforms_buffernode(bold_to_commonspace_transform_list, bold_to_commonspace_inverse_list, 
                commonspace_to_bold_transform_list, commonspace_to_bold_inverse_list,
                native_to_commonspace_transform_list, native_to_commonspace_inverse_list, 
                commonspace_to_native_transform_list, commonspace_to_native_inverse_list,
                bold_to_native_transform_list, bold_to_native_inverse_list):
        return
    transforms_buffernode_node = pe.Node(Function(input_names=output_transforms_between_spaces,
                                           output_names=[],
                                       function=transforms_buffernode),
                              name='transforms_buffernode')
    for target in output_transforms_between_spaces:
        workflow.connect([
            (prep_transforms_between_spaces_node, transforms_buffernode_node, [
                (target, target),
                ]),
            ])
    
    if not opts.bold_only:
        cross_modal_reg_wf = init_cross_modal_reg_wf(opts=opts)

        workflow.connect([
            (inputnode, cross_modal_reg_wf, [
                ('coreg_anat', 'inputnode.anat_ref'),
                ('coreg_mask', 'inputnode.anat_mask')]),
            (transitionnode, cross_modal_reg_wf, [
                ('corrected_EPI', 'inputnode.ref_bold_brain'),
                ('denoise_mask', 'inputnode.moving_mask'),
                ]),
            (cross_modal_reg_wf, outputnode, [
                ('outputnode.bold_to_anat_affine', 'bold_to_anat_affine'),
                ('outputnode.bold_to_anat_warp', 'bold_to_anat_warp'),
                ('outputnode.bold_to_anat_inverse_warp', 'bold_to_anat_inverse_warp'),
                ('outputnode.output_warped_bold', 'output_warped_bold'),
                ]),
            (cross_modal_reg_wf, prep_transforms_between_spaces_node, [
                ('outputnode.bold_to_anat_affine', 'bold_to_anat_affine'),
                ('outputnode.bold_to_anat_warp', 'bold_to_anat_warp'),
                ('outputnode.bold_to_anat_inverse_warp', 'bold_to_anat_inverse_warp'),
                ]),
            ])
        
    else:
        prep_transforms_between_spaces_node.inputs.bold_to_anat_warp = None
        prep_transforms_between_spaces_node.inputs.bold_to_anat_inverse_warp = None
        prep_transforms_between_spaces_node.inputs.bold_to_anat_affine = None

    # MAIN WORKFLOW STRUCTURE #######################################################
    workflow.connect([
        (inputnode, prep_transforms_between_spaces_node, [
            ('anat_to_commonspace_transform_list', 'anat_to_commonspace_transform_list'),
            ('anat_to_commonspace_inverse_list', 'anat_to_commonspace_inverse_list'),
            ('commonspace_to_anat_transform_list', 'commonspace_to_anat_transform_list'),
            ('commonspace_to_anat_inverse_list', 'commonspace_to_anat_inverse_list'),
            ]),
        (transitionnode, bold_stc_wf, [
            ('bold_file', 'inputnode.bold_file'),
            ]),
        (transitionnode, bold_hmc_wf, [
            ('bold_ref', 'inputnode.ref_image'),
            ]),
        (bold_hmc_wf, outputnode, [
            ('outputnode.motcorr_params', 'motcorr_params'),
            ('outputnode.hmc_qc_figure', 'hmc_qc_figure'),
            ('outputnode.hmc_qc_csv', 'hmc_qc_csv'),
            ('outputnode.hmc_qc_video', 'hmc_qc_video'),
            ]),
        (transitionnode, outputnode, [
            ('bold_ref', 'bold_ref'),
            ('init_denoise', 'init_denoise'),
            ('denoise_mask', 'denoise_mask'),
            ('corrected_EPI', 'corrected_EPI'),
            ]),
        (inputnode, estimate_motion_node, [
            ('bold', 'boldspace_bold'),
            ]),
        (bold_hmc_wf, estimate_motion_node, [
            ('outputnode.motcorr_params', 'motcorr_params'),
            ]),
        (estimate_motion_node, outputnode, [
            ('motion_params_csv', 'motion_params_csv'),
            ('FD_csv', 'FD_csv'),
            ]),
        ])

    if opts.log_transform:
        workflow.connect([
            (transitionnode, bold_hmc_wf, [
                ('log_bold', 'inputnode.bold_file')]),
            ])
    else:
        workflow.connect([
            (transitionnode, bold_hmc_wf, [
                ('bold_file', 'inputnode.bold_file')]),
            ])

    ####MANAGE GENERATING COMMON AND/OR NATIVESPACE OUTPUTS
    boldspace_brain_mask = pe.Node(ResampleMask(), name='boldspace_mask_resample')
    boldspace_brain_mask.inputs.name_suffix = 'boldspace_mask_resampled'
    boldspace_brain_mask.inputs.mask_file = str(opts.brain_mask)

    workflow.connect([
        (inputnode, boldspace_brain_mask, [
            ('bold', 'name_source'),
            ]),
        (transitionnode, boldspace_brain_mask, [
            ('bold_ref', 'ref_file'),
            ]),
        (prep_transforms_between_spaces_node, boldspace_brain_mask, [
            ('commonspace_to_bold_transform_list', 'transforms'),
            ('commonspace_to_bold_inverse_list', 'inverses'),
            ]),
        (boldspace_brain_mask, estimate_motion_node, [
            ('resampled_file', 'boldspace_brain_mask'),
            ]),
        (boldspace_brain_mask, outputnode, [
            ('resampled_file', 'boldspace_brain_mask')]),
    ])
        
    mask_commonspace_trans_wf = init_mask_preproc_trans_wf(opts=opts, name='mask_commonspace_trans_wf')
    mask_commonspace_trans_wf.inputs.inputnode.mask_transforms_list = []
    mask_commonspace_trans_wf.inputs.inputnode.mask_inverses = []
    workflow.connect([
        (inputnode, mask_commonspace_trans_wf, [
            ('commonspace_ref', 'inputnode.name_source'),
            ('commonspace_ref', 'inputnode.ref_file'),
            ]),
        (mask_commonspace_trans_wf, outputnode, [
            ('outputnode.brain_mask', 'commonspace_mask'),
            ('outputnode.WM_mask', 'commonspace_WM_mask'),
            ('outputnode.CSF_mask', 'commonspace_CSF_mask'),
            ('outputnode.vascular_mask', 'commonspace_vascular_mask'),
            ]),
        ])

    if opts.generate_nativespace:
        bold_native_trans_wf = init_bold_preproc_trans_wf(opts=opts, resampling_dim=opts.nativespace_resampling, name='bold_native_trans_wf')
        mask_native_trans_wf = init_mask_preproc_trans_wf(opts=opts, name='mask_native_trans_wf')

        if not opts.bold_only:
            if opts.bold_nativespace:
                workflow.connect([
                    (transitionnode, bold_native_trans_wf, [
                        ('bold_ref', 'inputnode.ref_file'),
                        ]),
                    ])
            else:
                workflow.connect([
                    (cross_modal_reg_wf, bold_native_trans_wf, [
                        ('outputnode.output_warped_bold', 'inputnode.ref_file'),
                        ]),
                    ])
        else:
            # for bold_only, the nativespace is basically the original EPI image space
            workflow.connect([
                (transitionnode, bold_native_trans_wf, [
                    ('bold_ref', 'inputnode.ref_file'),
                    ]),
                ])

        workflow.connect([
            (bold_stc_wf, bold_native_trans_wf, [
                ('outputnode.stc_file', 'inputnode.bold_file'),
                ]),
            (prep_transforms_between_spaces_node, bold_native_trans_wf, [
                ('bold_to_native_transform_list', 'inputnode.transforms_list'),
                ('bold_to_native_inverse_list', 'inputnode.inverses'),
                ]),
            (prep_transforms_between_spaces_node, mask_native_trans_wf, [
                ('commonspace_to_native_transform_list', 'inputnode.mask_transforms_list'),
                ('commonspace_to_native_inverse_list', 'inputnode.mask_inverses'),
                ]),
            (bold_hmc_wf, bold_native_trans_wf, [
                ('outputnode.motcorr_params', 'inputnode.motcorr_params')]),
            (inputnode, bold_native_trans_wf, [
                ('bold', 'inputnode.name_source'),
                ]),
            (inputnode, mask_native_trans_wf, [
                ('bold', 'inputnode.name_source'),
                ]),
            (bold_native_trans_wf, mask_native_trans_wf, [
                ('outputnode.bold_ref','inputnode.ref_file'),
                ]),
            (bold_native_trans_wf, outputnode, [
                ('outputnode.bold', 'native_bold'),
                ('outputnode.bold_ref','native_bold_ref'),
                ]),
            (mask_native_trans_wf, outputnode, [
                ('outputnode.brain_mask', 'native_brain_mask'),
                ('outputnode.WM_mask', 'native_WM_mask'),
                ('outputnode.CSF_mask', 'native_CSF_mask'),
                ('outputnode.vascular_mask', 'native_vascular_mask'),
                ]),
            ])

    if opts.generate_commonspace:
        bold_commonspace_trans_wf = init_bold_preproc_trans_wf(opts=opts, resampling_dim='ref_file', name='bold_commonspace_trans_wf')

        workflow.connect([
            (bold_stc_wf, bold_commonspace_trans_wf, [
                ('outputnode.stc_file', 'inputnode.bold_file')
                ]),
            (prep_transforms_between_spaces_node, bold_commonspace_trans_wf, [
                ('bold_to_commonspace_transform_list', 'inputnode.transforms_list'),
                ('bold_to_commonspace_inverse_list', 'inputnode.inverses'),
                ]),
            (bold_hmc_wf, bold_commonspace_trans_wf, [
            ('outputnode.motcorr_params', 'inputnode.motcorr_params')]),
            (inputnode, bold_commonspace_trans_wf, [
                ('bold', 'inputnode.name_source'),
                ('commonspace_ref', 'inputnode.ref_file'),
                ]),
            (bold_commonspace_trans_wf, outputnode, [
                ('outputnode.bold', 'commonspace_bold'),
                ]),
            ])

    return workflow



import os
import pathlib
from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu
from nipype.interfaces.io import DataSink
from nipype.interfaces.utility import Function
from rabies.utils import fill_split_dict, get_workflow_dict


def init_main_confound_correction_wf(preprocess_opts, cr_opts):
    from rabies.confound_correction_pkg.confound_correction import init_confound_correction_wf

    check_opts_compatibility(preprocess_opts, cr_opts)

    workflow = pe.Workflow(name='confound_correction_main_wf')

    preproc_output = os.path.abspath(str(cr_opts.preprocess_out))

    if cr_opts.read_datasink:
        split_dict, split_name, target_list = read_preproc_datasinks(preproc_output, nativespace_analysis=cr_opts.nativespace_analysis, preprocess_opts=preprocess_opts)
    else:
        split_dict, split_name, target_list = read_preproc_workflow(preproc_output, nativespace_analysis=cr_opts.nativespace_analysis, preprocess_opts=preprocess_opts)

    # filter inclusion/exclusion lists
    from rabies.utils import filter_scan_inclusion, filter_scan_exclusion
    split_name = filter_scan_inclusion(cr_opts.inclusion_ids, split_name)
    split_name = filter_scan_exclusion(cr_opts.exclusion_ids, split_name)

    number_functional_scans = len(split_name)
    if cr_opts.num_ITK_threads=='optimal': # override previous assignment from run_main(): an optimal thread number can be selected based on # of scans
        from nipype import logging
        log = logging.getLogger('nipype.workflow')
        num_ITK_threads=max(int(cr_opts.local_threads/number_functional_scans),1)
        log.info(f"An optimal number of {num_ITK_threads} ITK threads are allocated for {number_functional_scans} functional scans.")
        os.environ['RABIES_ITK_NUM_THREADS']=str(num_ITK_threads)
    else:
        num_ITK_threads = int(os.environ['RABIES_ITK_NUM_THREADS'])

    # setting up iterables from the BOLD scan splits
    main_split = pe.Node(niu.IdentityInterface(fields=['split_name']),
                         name="main_split")
    main_split.iterables = [('split_name', split_name)]

    # set output node from preprocessing
    def read_dict(split_dict, split_name, target_list):
        return [split_dict[split_name][target] for target in target_list]
    preproc_outputnode = pe.Node(Function(input_names=['split_dict', 'split_name', 'target_list'],
                                           output_names=target_list,
                                       function=read_dict),
                              name='preproc_outputnode')
    preproc_outputnode.inputs.split_dict = split_dict
    preproc_outputnode.inputs.target_list = target_list

    # need to set a buffer function which will be holding the preproc_outputnode outputs, 
    # so that it is saved in the workflow graph and can be read later during analysis
    def buffer_outputnode(input_bold=None, commonspace_bold=None, commonspace_mask=None, commonspace_WM_mask=None,
        commonspace_CSF_mask=None, commonspace_vascular_mask=None, motion_params_csv=None,
        FD_csv=None, commonspace_resampled_template=None, native_bold=None, native_bold_ref=None, 
        native_brain_mask=None, native_WM_mask=None, native_CSF_mask=None, native_vascular_mask=None,
        commonspace_to_native_transform_list=None, commonspace_to_native_inverse_list=None,
        native_to_commonspace_transform_list=None, native_to_commonspace_inverse_list=None):
        return
    buffer_outputnode_node = pe.Node(Function(input_names=target_list,
                                           output_names=[],
                                       function=buffer_outputnode),
                              name='buffer_outputnode')
    for target in target_list:
        workflow.connect([
            (preproc_outputnode, buffer_outputnode_node, [
                (target, target),
                ]),
            ])


    confound_correction_wf = init_confound_correction_wf(cr_opts=cr_opts)

    workflow.connect([
        (main_split, preproc_outputnode, [
            ("split_name", "split_name"),
            ]),
        (preproc_outputnode, confound_correction_wf, [
            ("motion_params_csv", "inputnode.motion_params_csv"),  # confounds file
            ("FD_csv", "inputnode.FD_file"),
            ("input_bold", "inputnode.raw_input_file"),
            ]),
        ])

    if cr_opts.nativespace_analysis:
        workflow.connect([
            (preproc_outputnode, confound_correction_wf, [
                ("native_bold", "inputnode.bold_file"),
                ("native_brain_mask", "inputnode.brain_mask"),
                ("native_WM_mask", "inputnode.WM_mask"),
                ("native_CSF_mask", "inputnode.CSF_mask"),
                ("native_vascular_mask", "inputnode.vascular_mask"),
                ]),
            ])
    else:
        workflow.connect([
            (preproc_outputnode, confound_correction_wf, [
                ("commonspace_bold", "inputnode.bold_file"),
                ("commonspace_mask", "inputnode.brain_mask"),
                ("commonspace_WM_mask", "inputnode.WM_mask"),
                ("commonspace_CSF_mask", "inputnode.CSF_mask"),
                ("commonspace_vascular_mask", "inputnode.vascular_mask"),
                ]),
            ])

    cr_output = os.path.abspath(str(cr_opts.output_dir))

    confound_correction_datasink = pe.Node(DataSink(base_directory=cr_output,
                                                    container="confound_correction_datasink"),
                                            name="confound_correction_datasink")
    workflow.connect([
        (confound_correction_wf, confound_correction_datasink, [
            ("outputnode.cleaned_path", "cleaned_timeseries"),
            ]),
        ])
    if cr_opts.ica_aroma['apply']:
        workflow.connect([
            (confound_correction_wf, confound_correction_datasink, [
                ("outputnode.aroma_out", "aroma_out"),
                ]),
            ])
    if cr_opts.frame_censoring['DVARS_censoring'] or cr_opts.frame_censoring['FD_censoring']:
        workflow.connect([
            (confound_correction_wf, confound_correction_datasink, [
                ("outputnode.frame_mask_file", "frame_censoring_mask"),
                ]),
            ])
        
    if cr_opts.generate_CR_null:
        plot_CR_overfit_node = pe.Node(Function(input_names=['mask_file', 'STD_file_path', 'CR_STD_file_path', 'random_CR_STD_file_path', 'corrected_CR_STD_file_path', 'figure_format'],
                                            output_names=['figure_path'],
                                        function=plot_CR_overfit),
                                name='plot_CR_overfit_node')
        plot_CR_overfit_node.inputs.figure_format = cr_opts.figure_format

        workflow.connect([
            (confound_correction_wf, plot_CR_overfit_node, [
                ("outputnode.STD_file", "STD_file_path"),
                ("outputnode.CR_STD_file", "CR_STD_file_path"),
                ("outputnode.random_CR_STD_file_path", "random_CR_STD_file_path"),
                ("outputnode.corrected_CR_STD_file_path", "corrected_CR_STD_file_path"),
                ]),
            (plot_CR_overfit_node, confound_correction_datasink, [
                ("figure_path", "plot_CR_overfit"),
                ]),
            ])
        
        if cr_opts.nativespace_analysis:
            workflow.connect([
                (preproc_outputnode, plot_CR_overfit_node, [
                    ("native_brain_mask", "mask_file"),
                    ]),
                ])
        else:
            workflow.connect([
                (preproc_outputnode, plot_CR_overfit_node, [
                    ("commonspace_mask", "mask_file"),
                    ]),
                ])

    if cr_opts.nativespace_analysis and cr_opts.resample_to_commonspace:
        from rabies.utils import ResampleVolumes
        cleaned_bold_to_commonspace_node = pe.Node(ResampleVolumes(
            resampling_dim='ref_file', interpolation=cr_opts.interpolation_sitk,
            rabies_data_type=cr_opts.data_type, apply_motcorr=False, clip_negative=False), 
            n_procs=num_ITK_threads,
            name='cleaned_bold_to_commonspace')
        workflow.connect([
            (confound_correction_wf, cleaned_bold_to_commonspace_node, [
                ("outputnode.cleaned_path", "in_file"),
                ("outputnode.cleaned_path", "name_source"),
                ]),
            (preproc_outputnode, cleaned_bold_to_commonspace_node, [
                ("native_to_commonspace_transform_list", "transforms_3d_files"),
                ("native_to_commonspace_inverse_list", "inverses_3d"),
                ("commonspace_resampled_template", "ref_file"),
                ]),
            (cleaned_bold_to_commonspace_node, confound_correction_datasink, [
                ("resampled_file", "cleaned_timeseries_commonspace"),
                ]),
            ])

    return workflow


def check_opts_compatibility(preprocess_opts, cr_opts):
    
    if preprocess_opts.CSF_mask is None:
        if cr_opts.ica_aroma['apply']:
            raise ValueError(
                'A CSF mask is necessary for AROMA, but none was retrieved from preprocessing.')
        if 'CSF_signal' in cr_opts.nuisance_regressors:
            raise ValueError(
                'CSF regression cannot be applied since no CSF mask was retrieved from preprocessing.')
        if 'aCompCor_5' in cr_opts.nuisance_regressors:
            raise ValueError(
                'aCompCor regression cannot be applied since no CSF mask was retrieved from preprocessing.')
        if 'aCompCor_percent' in cr_opts.nuisance_regressors:
            raise ValueError(
                'aCompCor regression cannot be applied since no CSF mask was retrieved from preprocessing.')

    if preprocess_opts.WM_mask is None:
        if 'WM_signal' in cr_opts.nuisance_regressors:
            raise ValueError(
                'WM regression cannot be applied since no WM mask was retrieved from preprocessing.')
        if 'aCompCor_5' in cr_opts.nuisance_regressors:
            raise ValueError(
                'aCompCor regression cannot be applied since no WM mask was retrieved from preprocessing.')
        if 'aCompCor_percent' in cr_opts.nuisance_regressors:
            raise ValueError(
                'aCompCor regression cannot be applied since no WM mask was retrieved from preprocessing.')
        
    if preprocess_opts.vascular_mask is None:
        if 'vascular_signal' in cr_opts.nuisance_regressors:
            raise ValueError(
                'vascular regression cannot be applied since no vascular mask was retrieved from preprocessing.')
        
    if cr_opts.nativespace_analysis:
        if not preprocess_opts.generate_nativespace: # nativespace timeseries are needed
            raise ValueError(
                'Timeseries in nativespace were not generated during preprocessing.')
    else:
        if not preprocess_opts.generate_commonspace: # commonspace timeseries are needed
            raise ValueError(
                'Timeseries in commonspace were not generated during preprocessing.')


def read_preproc_datasinks(preproc_output, nativespace_analysis, preprocess_opts):
    import pathlib
    import glob

    fast_commonspace=preprocess_opts.commonspace_reg['fast_commonspace']
    atlas_reg_script=preprocess_opts.commonspace_reg['template_registration']

    template_file = glob.glob(f'{preproc_output}/bold_datasink/commonspace_resampled_template/*')
    if len(template_file)==1:
        template_file = template_file[0]
    else:
        raise ValueError(f"Multiple files were found in {preproc_output}/bold_datasink/commonspace_resampled_template/"
                        "but there should only be one template file.")

    split_dict = {}
    bold_scan_list = get_files_from_tree(f'{preproc_output}/bold_datasink/input_bold')
    split_name = []
    for f in bold_scan_list:
        name = pathlib.Path(f).name.rsplit(".nii")[0]
        split_name.append(name)
        split_dict[name]={}
        split_dict[name]['commonspace_resampled_template']=template_file

    directory_list = [['bold_datasink','input_bold'],['bold_datasink','commonspace_mask'],
        ['motion_datasink','motion_params_csv'], ['motion_datasink','FD_csv']]

    # these parameters may be empty
    for opt_key in ['WM_mask','CSF_mask','vascular_mask']:
        opt_file = getattr(preprocess_opts, opt_key)
        if opt_file is not None:
            directory_list.append(['bold_datasink',f'commonspace_{opt_key}'])
        else:
            directory_list.append([None,f'commonspace_{opt_key}'])

    if not nativespace_analysis:
        if not preprocess_opts.generate_commonspace: # nativespace timeseries are needed
            raise ValueError(
                'Timeseries in commonspace were not generated during preprocessing.')

        directory_list+=[['bold_datasink','commonspace_bold']]
    else:
        if not preprocess_opts.generate_nativespace: # nativespace timeseries are needed
            raise ValueError(
                'Timeseries in nativespace were not generated during preprocessing.')
        
        directory_list+=[['bold_datasink','native_bold'], ['bold_datasink','native_brain_mask'], ['bold_datasink', 'native_bold_ref']]
        # these parameters may be empty
        for opt_key in ['WM_mask','CSF_mask','vascular_mask']:
            opt_file = getattr(preprocess_opts, opt_key)
            if opt_file is not None:
                directory_list.append(['bold_datasink',f'native_{opt_key}'])
            else:
                directory_list.append([None,f'native_{opt_key}'])

    target_list=['commonspace_resampled_template']
    for datasink,target in directory_list:
        target_list.append(target)
        if datasink is None:
            for split in split_name:
                split_dict[split][target]=None
        else:
            if not os.path.isdir(f'{preproc_output}/{datasink}/{target}'):
                raise ValueError(f"The directory {preproc_output}/{datasink}/{target} does not exist. Make sure that all required "
                    "datasink outputs are available.")
            file_list = get_files_from_tree(f'{preproc_output}/{datasink}/{target}')
            if target in ['commonspace_mask','commonspace_WM_mask','commonspace_CSF_mask','commonspace_vascular_mask']: # the split_name is not found in the filename for commonspace masks
                if len(file_list)>1:
                    raise ValueError(f"There is more than one {target} file found: {file_list}")
                for split in split_name:
                    split_dict[split][target]=file_list[0]

            for f in file_list:
                for split in split_name:
                    if split in f:
                        split_dict[split][target]=f
                        break

    if nativespace_analysis:
        ###
        # For the transforms, there needs to be a different file matching, where files may be named based on the anat
        # scan, so here we match the BIDS specs. The transforms to native space are put together into a prepared list of transforms.
        ###
        if fast_commonspace:
            directory_list+=[['transforms_datasink','anat_to_atlas_affine']]
            if atlas_reg_script=='SyN':
                directory_list+=[['transforms_datasink','anat_to_atlas_inverse_warp']]
                directory_list+=[['transforms_datasink','anat_to_atlas_warp']]
        else:
            directory_list+=[['transforms_datasink','unbiased_to_atlas_affine'], 
                ['transforms_datasink','anat_to_unbiased_affine'], ['transforms_datasink','anat_to_unbiased_inverse_warp'],
                ['transforms_datasink','anat_to_unbiased_warp']]
            if atlas_reg_script=='SyN':
                directory_list+=[['transforms_datasink','unbiased_to_atlas_inverse_warp']]
                directory_list+=[['transforms_datasink','unbiased_to_atlas_warp']]
        if not preprocess_opts.bold_only:
            directory_list+=[['transforms_datasink','bold_to_anat_affine']]
            if preprocess_opts.bold2anat_coreg['registration']=='SyN':
                directory_list+=[['transforms_datasink','bold_to_anat_inverse_warp']]
                directory_list+=[['transforms_datasink','bold_to_anat_warp']]
            
        from bids.layout import parse_file_entities
        for datasink,target in directory_list:
            if datasink is None:
                continue # None values are attributed at the end

            if not os.path.isdir(f'{preproc_output}/{datasink}/{target}'):
                raise ValueError(f"The directory {preproc_output}/{datasink}/{target} does not exist. Make sure that all required "
                    "datasink outputs are available.")
            target_list.append(target)
            file_list = get_files_from_tree(f'{preproc_output}/{datasink}/{target}')
            for split in split_name:
                # for the unbiased to atlas transforms it is not subject-specific
                if target in ['unbiased_to_atlas_affine', 'unbiased_to_atlas_inverse_warp', 'unbiased_to_atlas_warp']:
                    if not len(file_list)==1:
                        raise ValueError(f"There should be only a single transform from unbiased to atlas space. Instead there is {file_list}.")
                    split_dict[split][target]=file_list[0]
                else:
                    parser_split = parse_file_entities('/'+split)
                    for f in file_list:
                        parser_f = parse_file_entities(f)
                        # subject info is mandatory
                        if not 'subject' in list(parser_f.keys()):
                            raise ValueError(f"The file {f} is missing the 'subject' BIDS specs.")
                        if not 'subject' in list(parser_split.keys()):
                            raise ValueError(f"The file {split} is missing the 'subject' BIDS specs.")
                        if parser_f['subject']==parser_split['subject']:
                            # if session was specified in the BOLD file, it should also be in the target file
                            if 'session' in list(parser_split.keys()):
                                if not 'session' in list(parser_f.keys()):
                                    raise ValueError(f"The file {f} is missing the 'session' BIDS specs.")
                                if parser_f['session']==parser_split['session']:
                                    split_dict[split][target]=f
                                    break
                            else:
                                split_dict[split][target]=f
                                break

        for split in split_name:
            if fast_commonspace:
                anat_to_atlas_affine = split_dict[split]['anat_to_atlas_affine']
                if atlas_reg_script=='SyN':
                    anat_to_atlas_inverse_warp = split_dict[split]['anat_to_atlas_inverse_warp']
                    anat_to_atlas_warp = split_dict[split]['anat_to_atlas_warp']
                    commonspace_to_anat_transform_list=[anat_to_atlas_affine,anat_to_atlas_inverse_warp]
                    commonspace_to_anat_inverse_list=[1,0]
                    anat_to_commonspace_transform_list=[anat_to_atlas_warp,anat_to_atlas_affine]
                    anat_to_commonspace_inverse_list=[0,0]
                else:
                    commonspace_to_anat_transform_list=[anat_to_atlas_affine]
                    commonspace_to_anat_inverse_list=[1]
                    anat_to_commonspace_transform_list=[anat_to_atlas_affine]
                    anat_to_commonspace_inverse_list=[0]
            else:
                anat_to_unbiased_inverse_warp = split_dict[split]['anat_to_unbiased_inverse_warp']
                anat_to_unbiased_warp = split_dict[split]['anat_to_unbiased_warp']
                anat_to_unbiased_affine = split_dict[split]['anat_to_unbiased_affine']
                unbiased_to_atlas_affine = split_dict[split]['unbiased_to_atlas_affine']
                if atlas_reg_script=='SyN':
                    unbiased_to_atlas_inverse_warp = split_dict[split]['unbiased_to_atlas_inverse_warp']
                    unbiased_to_atlas_warp = split_dict[split]['unbiased_to_atlas_warp']
                    commonspace_to_anat_transform_list=[anat_to_unbiased_affine,anat_to_unbiased_inverse_warp,unbiased_to_atlas_affine,unbiased_to_atlas_inverse_warp]
                    commonspace_to_anat_inverse_list=[1,0,1,0]
                    anat_to_commonspace_transform_list=[unbiased_to_atlas_warp,unbiased_to_atlas_affine,anat_to_unbiased_warp,anat_to_unbiased_affine]
                    anat_to_commonspace_inverse_list=[0,0,0,0]
                else:
                    commonspace_to_anat_transform_list=[anat_to_unbiased_affine,anat_to_unbiased_inverse_warp,unbiased_to_atlas_affine]
                    commonspace_to_anat_inverse_list=[1,0,1]
                    anat_to_commonspace_transform_list=[unbiased_to_atlas_affine,anat_to_unbiased_warp,anat_to_unbiased_affine]
                    anat_to_commonspace_inverse_list=[0,0,0]
            if preprocess_opts.bold_only:
                bold_to_anat_warp = None
                bold_to_anat_inverse_warp = None
                bold_to_anat_affine = None
            else:
                bold_to_anat_affine = split_dict[split]['bold_to_anat_affine']
                if preprocess_opts.bold2anat_coreg['registration']=='SyN':
                    bold_to_anat_inverse_warp = split_dict[split]['bold_to_anat_inverse_warp']
                    bold_to_anat_warp = split_dict[split]['bold_to_anat_warp']
                else:
                    bold_to_anat_inverse_warp = 'NULL'
                    bold_to_anat_warp = 'NULL'

            from rabies.preprocess_pkg.bold_main_wf import prep_transforms_between_spaces
            [bold_to_commonspace_transform_list, bold_to_commonspace_inverse_list, 
            commonspace_to_bold_transform_list, commonspace_to_bold_inverse_list,
            native_to_commonspace_transform_list, native_to_commonspace_inverse_list, 
            commonspace_to_native_transform_list, commonspace_to_native_inverse_list,
            bold_to_native_transform_list, bold_to_native_inverse_list] = prep_transforms_between_spaces(anat_to_commonspace_transform_list,anat_to_commonspace_inverse_list, bold_to_anat_warp, bold_to_anat_inverse_warp, bold_to_anat_affine, 
                                commonspace_to_anat_transform_list, commonspace_to_anat_inverse_list, bold_only=preprocess_opts.bold_only, bold_nativespace=preprocess_opts.bold_nativespace)

            split_dict[split]['commonspace_to_native_transform_list'] = commonspace_to_native_transform_list
            split_dict[split]['commonspace_to_native_inverse_list'] = commonspace_to_native_inverse_list
            split_dict[split]['native_to_commonspace_transform_list'] = native_to_commonspace_transform_list
            split_dict[split]['native_to_commonspace_inverse_list'] = native_to_commonspace_inverse_list
        target_list += ['commonspace_to_native_transform_list', 'commonspace_to_native_inverse_list']
        target_list += ['native_to_commonspace_transform_list', 'native_to_commonspace_inverse_list']

        if fast_commonspace:
            if atlas_reg_script=='SyN':
                target_list.remove('anat_to_atlas_inverse_warp')
                target_list.remove('anat_to_atlas_warp')
            target_list.remove('anat_to_atlas_affine')
        else:
            target_list.remove('anat_to_unbiased_inverse_warp')
            target_list.remove('anat_to_unbiased_warp')
            target_list.remove('anat_to_unbiased_affine')
            if atlas_reg_script=='SyN':
                target_list.remove('unbiased_to_atlas_inverse_warp')
                target_list.remove('unbiased_to_atlas_warp')
            target_list.remove('unbiased_to_atlas_affine')

    return split_dict, split_name, target_list


def get_files_from_tree(startpath):
    file_list=[]
    for root, dirs, files in os.walk(startpath):
        for f in files:
            file_list.append(f'{root}/{f}')
    return file_list


def read_preproc_workflow(preproc_output, nativespace_analysis, preprocess_opts):

    preproc_workflow_file = f'{preproc_output}/rabies_preprocess_workflow.pkl'

    node_dict = get_workflow_dict(preproc_workflow_file)

    match_targets = {'input_bold':['main_wf.input_bold', 'selected_file'],
                    'commonspace_mask':['main_wf.bold_main_wf.mask_commonspace_trans_wf.brain_mask_resample', 'resampled_file'],
                    'motion_params_csv':['main_wf.bold_main_wf.estimate_motion_node', 'motion_params_csv'],
                    'FD_csv':['main_wf.bold_main_wf.estimate_motion_node', 'FD_csv'],
                    'commonspace_resampled_template':['main_wf.resample_template', 'commonspace_template'],
                    'commonspace_to_native_transform_list':['main_wf.bold_main_wf.prep_transforms_between_spaces', 'commonspace_to_native_transform_list'],
                    'commonspace_to_native_inverse_list':['main_wf.bold_main_wf.prep_transforms_between_spaces', 'commonspace_to_native_inverse_list'],
                    'native_to_commonspace_transform_list':['main_wf.bold_main_wf.prep_transforms_between_spaces', 'native_to_commonspace_transform_list'],
                    'native_to_commonspace_inverse_list':['main_wf.bold_main_wf.prep_transforms_between_spaces', 'native_to_commonspace_inverse_list'],
                    }

    # these parameters may be empty
    for opt_key in ['WM_mask','CSF_mask','vascular_mask']:
        opt_file = getattr(preprocess_opts, opt_key)
        if opt_file is not None:
            match_targets[f'commonspace_{opt_key}'] = [f'main_wf.bold_main_wf.mask_commonspace_trans_wf.{opt_key}_resample', 'resampled_file']
        else: # None values are propagated
            match_targets[f'commonspace_{opt_key}'] = None

    if not nativespace_analysis:
        if not preprocess_opts.generate_commonspace: # nativespace timeseries are needed
            raise ValueError(
                'Timeseries in commonspace were not generated during preprocessing.')
        match_targets.update({
                        'commonspace_bold':['main_wf.bold_main_wf.bold_commonspace_trans_wf.bold_transform', 'resampled_file'],
                        })
    else:    
        if not preprocess_opts.generate_nativespace: # nativespace timeseries are needed
            raise ValueError(
                'Timeseries in nativespace were not generated during preprocessing.')

        match_targets.update({'native_bold':['main_wf.bold_main_wf.bold_native_trans_wf.bold_transform', 'resampled_file'],
                        'native_bold_ref':['main_wf.bold_main_wf.bold_native_trans_wf.gen_bold_ref.gen_ref', 'ref_image'],                              
                        'native_brain_mask':['main_wf.bold_main_wf.mask_native_trans_wf.brain_mask_resample', 'resampled_file'],
                        })
        # these parameters may be empty
        for opt_key in ['WM_mask','CSF_mask','vascular_mask']:
            opt_file = getattr(preprocess_opts, opt_key)
            if opt_file is not None:
                match_targets[f'native_{opt_key}'] = [f'main_wf.bold_main_wf.mask_native_trans_wf.{opt_key}_resample', 'resampled_file']
            else: # None values are propagated
                match_targets[f'native_{opt_key}'] = None

    split_dict = {}
    split_name = []
    # preparing a new iterative node where each BOLD scan is a different split
    [unit_bold, output_bold] = match_targets['input_bold']
    bold_dict = node_dict[unit_bold]
    # fill each BOLD scan split with proper affiliated outputs from preprocessing
    fill_split_dict(bold_dict, output_bold, split_name, split_dict, [], node_dict, match_targets)

    target_list = list(match_targets.keys())

    return split_dict, split_name, target_list


def plot_CR_overfit(mask_file, STD_file_path, CR_STD_file_path, random_CR_STD_file_path, corrected_CR_STD_file_path, figure_format):

    for file in STD_file_path, CR_STD_file_path, random_CR_STD_file_path, corrected_CR_STD_file_path:
        if 'empty' in file:
            return file

    import os
    import matplotlib.pyplot as plt
    import SimpleITK as sitk
    nrows = 4
    fig, axes = plt.subplots(nrows=nrows, ncols=3, figsize=(12*3, 2*nrows))
    plt.tight_layout()

    from rabies.visualization import plot_3d
    volume_indices = sitk.GetArrayFromImage(sitk.ReadImage(mask_file, sitk.sitkFloat32)).astype(bool)

    def get_vmax(sitk_img, volume_indices):
        # select vmax at 95th percentile value
        vector = sitk.GetArrayFromImage(sitk_img)[volume_indices].flatten()
        vector.sort()
        vmax = vector[int(len(vector)*0.95)]
        return vmax
    
    title_list = ['$\mathregular{BOLD_{SD}}$', '$\mathregular{CR_{SD}}$', 
                  '$\mathregular{CR_{SD}}$ random', '$\mathregular{CR_{SD}}$ corrected']
    row=0
    for file, title in zip([STD_file_path, CR_STD_file_path, random_CR_STD_file_path, corrected_CR_STD_file_path], title_list):
        sitk_img = sitk.ReadImage(file)
        cbar_list = plot_3d(axes[row, :], sitk_img, fig, vmin=0, vmax=get_vmax(sitk_img, volume_indices),
                    cmap='inferno', alpha=1, cbar=True, num_slices=6)
        for cbar in cbar_list:
            cbar.ax.get_yaxis().labelpad = 35
            cbar.set_label('Standard \n Deviation', fontsize=17, rotation=270, color='white')
            cbar.ax.tick_params(labelsize=15)
        for ax in axes[row, :]:
            ax.set_title(title, fontsize=30, color='white')
        row +=1

    import pathlib
    filename_template = pathlib.Path(STD_file_path).name.rsplit("_STD_map.nii.gz")[0]
    figure_path = os.path.abspath(filename_template)+f'_CR_overfit.{figure_format}'
    fig.savefig(figure_path, bbox_inches='tight')

    return figure_path




#!/usr/bin/env python

# This file generates steps of registration between two images and attempts to compensate
# For ANTs' dependency on the resolution of the file

# We do this by defining two scales to step over
# blur_scale, which is the real-space steps in blurring we will do
# shrink_scale, which is the subsampling scale that is 1/2 the fwhm blur scale, adjusted for file minimum resolution and max size

from __future__ import division, print_function

import argparse
import sys
import re


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)

def check_positive(value):
    ivalue = int(value)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError("%s is an invalid positive int value" % value)
    return ivalue

class SplitArgsComma(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values.rstrip(',').split(','))

parser.add_argument(
    '--min', help='minimum resolution of fixed image (mm)', type=float, required=True)
parser.add_argument(
    '--max', help='maximum dimension of fixed image features (mm)', type=float, required=True)
parser.add_argument(
    '--start-scale', help='set starting sigma blur scale (mm), default calculated from max size', type=float)
parser.add_argument(
    '--final-iterations', help='total number of iterations at lowest scale', type=int, default=20)
parser.add_argument(
    '--output', help='type of output to generate', default='generic',
      choices=['generic', 'affine', 'modelbuild', 'twolevel_dbm', 'multilevel-halving', 'exhaustive-affine',
                'lsq6', 'lsq9', 'lsq12', 'rigid', 'similarity','volgenmodel','affine-plain'])
parser.add_argument('--volgen-iteration', help='for volgenmodel mode, control which iteration is output', default=0, type=int)
parser.add_argument('--step-size', help='step size for fwhm scale space, default 1/2 of voxel size', type=float)
parser.add_argument(
    '--convergence', help='set convergence for generated stages', default='1e-6')
parser.add_argument(
    '--convergence-window', help='set convergence window for generated stages', default='10')
parser.add_argument(
    '--close', help='images are already close, skip large scales of pyramid and translation and rigid steps for affine', action='store_true')
parser.add_argument(
    '--rough', help='skip full-resolution iterations for a "rough" alignment', action='store_true')
parser.add_argument(
    '--affine-metric', help='which metric to use for affine stages, use comma separated list for multiple image pairs (MI, Mattes, GC, MeanSquares, Demons)', default='Mattes')
parser.add_argument('--reg-pairs', help='number of image pairs for affine output', default=1, type=check_positive)
parser.add_argument('--reg-pairs-weights', help='either a single number for all weights, or a comma separated list of weights equal to reg_pairs', default=['1'], action=SplitArgsComma)
parser.add_argument('--no-masks', help='for linear registration outputs skip repeat stages with masks', action='store_true')

parser.add_argument('--override-shrink-factors', help='override calculation of optimal image pyramid with specified settings')
parser.add_argument('--override-smoothing-sigmas', help='override calculation of optimal image pyramid with specified settings')
parser.add_argument('--override-convergence', help='override calculation of optimal image pyramid with specified settings')


args = parser.parse_args()

# Setup inital inputs
min_resolution = args.min
max_size = args.max

affinemetric = args.affine_metric.rstrip(',').split(",")
affineweights= args.reg_pairs_weights

if len(affinemetric) != args.reg_pairs:
  if len(affinemetric) != 1:
    sys.exit("Number of metrics provided not equal to the number of image pairs, or 1")
  else:
    affinemetric = affinemetric * args.reg_pairs

if len(args.reg_pairs_weights) != args.reg_pairs:
  if len(args.reg_pairs_weights) != 1:
    sys.exit("Number of affine weights provided not equal to the number of image pairs, or 1")
  else:
    affineweights = affineweights * args.reg_pairs


if (args.output in ['affine', 'multilevel-halving', 'exhaustive-affine','lsq6', 'lsq9', 'lsq12', 'rigid', 'similarity']):
  if args.step_size:
     step_size = args.step_size
  else:
    step_size = min_resolution / 2
  if args.start_scale:
     starting_sigma = args.start_scale
  else:
    starting_sigma = max_size / 32
else:
  if args.step_size:
     step_size = args.step_size
  else:
    step_size = min_resolution / 2

  if args.start_scale:
     starting_sigma = args.start_scale
  else:
    starting_sigma = max_size / 12.3125 / 4

if args.close:
  starting_sigma = starting_sigma / 2

# Step down sigmas by step_size from starting_sigma to zero
sigmas = [ x * step_size for x in list(range(int(round(starting_sigma/step_size)),-1,-1)) ]

# Shrinks are 2x sigma, with a capped maximum shrink
shrinks = [ round(max(min(max_size/min_resolution/32, 2*x/min_resolution),1)) for x in sigmas ]

iterations = [ min(500,int(args.final_iterations * 3**(max(0,x - 1)))) for x in shrinks ]

if args.rough:
  to_remove = [i for i in range(len(shrinks)) if shrinks[i]==1]
  sigmas = [i for j, i in enumerate(sigmas) if j not in to_remove]
  shrinks = [i for j, i in enumerate(shrinks) if j not in to_remove]
  iterations = [i for j, i in enumerate(iterations) if j not in to_remove]
  sigmas.append('0')
  shrinks.append('1')
  iterations.append('0')

# Convert to strings
sigmas = [ str(x) for x in sigmas ]
shrinks = [ str(int(x)) for x in shrinks ]
iterations = [ str(x) for x in iterations ]

suffix = "mm"

if args.override_shrink_factors and args.override_smoothing_sigmas and args.override_convergence:
   if re.search("mm", args.override_smoothing_sigmas):
      suffix="mm"
      sigmas = args.override_smoothing_sigmas.strip("mm").split("x")
   elif re.search("vox", args.override_smoothing_sigmas):
      suffix="vox"
      sigmas = args.override_smoothing_sigmas.strip("vox").split("x")
   else:
      sigmas = args.override_smoothing_sigmas.split("x")
      suffix = ""
   shrinks = args.override_shrink_factors.split("x")
   iterations = args.override_convergence.split("x")

# Setup transforms
if args.output in ["multilevel-halving", "affine", "lsq12","exhaustive-affine"]:
  transforms = ["--transform Translation[ ",
                "--transform Rigid[ ",
                "--transform Similarity[ ",
                "--transform Affine[ "]
elif args.output in ["lsq9","similarity"]:
  transforms = ["--transform Translation[ ",
                "--transform Rigid[ ",
                "--transform Similarity[ ",
                "--transform Similarity[ "]
elif args.output in ["lsq6","rigid"]:
  transforms = ["--transform Translation[ ",
                "--transform Rigid[ ",
                "--transform Rigid[ ",
                "--transform Rigid[ "]
elif args.output in ['affine-plain']:
  transforms = ["--transform Rigid[ ",
                "--transform Similarity[ ",
                "--transform Affine[ "]
if not args.close:
  gradient_steps = [ 0.5, 0.33437015, 0.2236068, 0.1 ]
  gradient_steps_repeat = [ 0.5, 0.33437015, 0.14953488, 0.1 ]
else:
  gradient_steps = [ 0.1, 0.1, 0.1, 0.1 ]
  gradient_steps_repeat = [ 0.1, 0.1, 0.1, 0.1 ]

masks = ["--masks [ NOMASK,NOMASK ]",
         "--masks [ NOMASK,NOMASK ]",
         "--masks [ NOMASK,NOMASK ]",
         "--masks [ ${fixedmask},${movingmask} ]" ]

# If defined, repeat a stage previously done with a mask
repeatmask = [ False,
               False,
               "--masks [ ${fixedmask},${movingmask} ]",
               False ]

if args.output == 'exhaustive-affine' or args.output == 'affine-plain':
    for i, transform in enumerate(transforms):
            if i == len(transforms) - 1:
              print(transform + str(gradient_steps[i]) + " ]", end=' \\\n')
              for j in range(1, args.reg_pairs+1):
                  print("\t--metric {affinemetric}[ ${{fixedfile{j}}},${{movingfile{j}}},{affineweights},32,None,1,1 ]".format(j=j, affinemetric=affinemetric[j-1], affineweights=affineweights[j-1]), end=' \\\n')
              print("\t--convergence [ {},{},{} ]".format("x".join(iterations), args.convergence, args.convergence_window), end=' \\\n')
              print("\t--shrink-factors {}".format("x".join(shrinks)), end=' \\\n')
              if args.no_masks:
                print("\t--smoothing-sigmas {}{}".format("x".join(sigmas),suffix), end=' ')
              else:
                print("\t--smoothing-sigmas {}{}".format("x".join(sigmas),suffix), end=' \\\n')
                print("\t" + masks[i], end=' ')
            else:
              print(transform + str(gradient_steps[i]) + " ]", end=' \\\n')
              for j in range(1, args.reg_pairs+1):
                  print("\t--metric {affinemetric}[ ${{fixedfile{j}}},${{movingfile{j}}},{affineweights},32,None,1,1 ]".format(j=j, affinemetric=affinemetric[j-1], affineweights=affineweights[j-1]), end=' \\\n')
              print("\t--convergence [ {},{},{} ]".format("x".join(iterations), args.convergence, args.convergence_window), end=' \\\n')
              print("\t--shrink-factors {}".format("x".join(shrinks)), end=' \\\n')
              print("\t--smoothing-sigmas {}{}".format("x".join(sigmas),suffix), end=' \\\n')
              if not args.no_masks:
                print("\t" + masks[i], end=' \\\n')
                if repeatmask[i]:
                  print(transform + str(gradient_steps[i]) + " ]", end=' \\\n')
                  for j in range(1, args.reg_pairs+1):
                    print("\t--metric {affinemetric}[ ${{fixedfile{j}}},${{movingfile{j}}},{affineweights},32,None,1,1 ]".format(j=j, affinemetric=affinemetric[j-1], affineweights=affineweights[j-1]), end=' \\\n')
                  print("\t--convergence [ {},{},{} ]".format("x".join(iterations), args.convergence, args.convergence_window), end=' \\\n')
                  print("\t--shrink-factors {}".format("x".join(shrinks)), end=' \\\n')
                  print("\t--smoothing-sigmas {}{}".format("x".join(sigmas),suffix), end=' \\\n')
                  print("\t" + repeatmask[i], end=' \\\n')

elif args.output == 'twolevel_dbm':
    print("--reg-iterations {}".format("x".join(iterations)), end=' \\\n')
    print("--reg-shrinks {}".format("x".join(shrinks)), end=' \\\n')
    print("--reg-smoothing {}{}".format("x".join(sigmas),suffix), end=' ')

elif args.output == 'modelbuild':
    print("-q {}".format("x".join(iterations)), end=' \\\n')
    print("-f {}".format("x".join(shrinks)), end=' \\\n')
    print("-s {}{}".format("x".join(sigmas),suffix), end=' ')

elif args.output == 'generic':
      print("--convergence [ {},{},{} ]".format("x".join(iterations), args.convergence, args.convergence_window), end=' \\\n')
      print("--shrink-factors {}".format("x".join(shrinks)), end=' \\\n')
      print("--smoothing-sigmas {}{}".format("x".join(sigmas),suffix), end=' ')

elif args.output == 'volgenmodel':
    print("--convergence [ {}x0,{},{} ]".format("x".join(iterations[0:min(args.volgen_iteration+1,len(iterations))]), args.convergence, args.convergence_window), end=' \\\n')
    print("--shrink-factors {}x1".format("x".join(shrinks[0:min(args.volgen_iteration+1,len(shrinks))])), end=' \\\n')
    print("--smoothing-sigmas {}x0{}".format("x".join(sigmas[0:min(args.volgen_iteration+1,len(sigmas))]),suffix), end=' ')

else:

    slicestart = [ 0,
                   int(round(0.25*len(sigmas))),
                   int(round(0.50*len(sigmas))),
                   int(round(0.75*len(sigmas)))]

    sliceend = [   int(round(0.5*len(sigmas))),
                   int(round(0.75*len(sigmas))),
                   -2,
                   -1]

    if args.close:
      transforms = transforms[2:]
      slicestart = slicestart[0:3]
      sliceend = sliceend[1:]
      masks = masks[2:]
      repeatmask = repeatmask[2:]


    for i, transform in enumerate(transforms):
      if i == len(transforms) - 1:
        print(transform + str(gradient_steps[i]) + " ]", end=' \\\n')
        for j in range(1, args.reg_pairs+1):
          print("\t--metric {affinemetric}[ ${{fixedfile{j}}},${{movingfile{j}}},{affineweights},64,None,1,1 ]".format(j=j, affinemetric=affinemetric[j-1], affineweights=affineweights[j-1]), end=' \\\n')
        print("\t--convergence [ {},{},{} ]".format("x".join(iterations[slicestart[i]:]), args.convergence, args.convergence_window), end=' \\\n')
        print("\t--shrink-factors {}".format("x".join(shrinks[slicestart[i]:])), end=' \\\n')
        if args.no_masks:
          print("\t--smoothing-sigmas {}{}".format("x".join(sigmas[slicestart[i]:]),suffix), end=' ')
        else:
          print("\t--smoothing-sigmas {}{}".format("x".join(sigmas[slicestart[i]:]),suffix), end=' \\\n')
          print("\t" + masks[i], end=' ')
      else:
        print(transform + str(gradient_steps[i]) + " ]", end=' \\\n')
        for j in range(1, args.reg_pairs+1):
          print("\t--metric {affinemetric}[ ${{fixedfile{j}}},${{movingfile{j}}},{affineweights},32,None,1,1 ]".format(j=j, affinemetric=affinemetric[j-1], affineweights=affineweights[j-1]), end=' \\\n')
        print("\t--convergence [ {},{},{} ]".format("x".join(iterations[slicestart[i]:max(-slicestart[i]+1,sliceend[i])]), args.convergence, args.convergence_window), end=' \\\n')
        print("\t--shrink-factors {}".format("x".join(shrinks[slicestart[i]:max(-slicestart[i]+1,sliceend[i])])), end=' \\\n')
        print("\t--smoothing-sigmas {}{}".format("x".join(sigmas[slicestart[i]:max(-slicestart[i]+1,sliceend[i])]),suffix), end=' \\\n')
        if not args.no_masks:
          print("\t" + masks[i], end=' \\\n')
          if repeatmask[i]:
            print(transform + str(gradient_steps_repeat[i]) + " ]", end=' \\\n')
            for j in range(1, args.reg_pairs+1):
                print("\t--metric {affinemetric}[ ${{fixedfile{j}}},${{movingfile{j}}},{affineweights},32,None,1,1 ]".format(j=j, affinemetric=affinemetric[j-1], affineweights=affineweights[j-1]), end=' \\\n')
            print("\t--convergence [ {},{},{} ]".format("x".join(iterations[slicestart[i]:max(-slicestart[i]+1,sliceend[i])]), args.convergence, args.convergence_window), end=' \\\n')
            print("\t--shrink-factors {}".format("x".join(shrinks[slicestart[i]:max(-slicestart[i]+1,sliceend[i])])), end=' \\\n')
            print("\t--smoothing-sigmas {}{}".format("x".join(sigmas[slicestart[i]:max(-slicestart[i]+1,sliceend[i])]),suffix), end=' \\\n')
            print("\t" + repeatmask[i], end=' \\\n')



def compute_edge_mask(mask_array, num_edge_voxels=1):
    import numpy as np
    #custom function for computing edge mask from an input brain mask
    shape = mask_array.shape

    #iterate through all voxels from the three dimensions and look if it contains surrounding voxels
    edge_mask = np.zeros(shape, dtype=bool)
    num_voxel = 0
    while num_voxel < num_edge_voxels:
        for x in range(shape[0]):
            for y in range(shape[1]):
                for z in range(shape[2]):
                    #only look if the voxel is part of the mask
                    if mask_array[x, y, z]:
                        if (mask_array[x-1:x+2, y-1:y+2, z-1:z+2] == 0).sum() > 0:
                            edge_mask[x, y, z] = 1
        mask_array = mask_array-edge_mask
        num_voxel += 1

    return edge_mask


import os
import pickle
import SimpleITK as sitk
from nipype import logging, config
import pathlib
from .boilerplate import *
from .parser import get_parser,read_parser
from .preprocess_pkg.utils import convert_to_RAS

# setting all default template files
if 'XDG_DATA_HOME' in os.environ.keys():
    rabies_path = os.environ['XDG_DATA_HOME']+'/rabies'
else:
    rabies_path = os.environ['HOME']+'/.local/share/rabies'

DSURQE_ANAT=f"{rabies_path}/DSURQE_40micron_average.nii.gz"
DSURQE_MASK=f"{rabies_path}/DSURQE_40micron_mask.nii.gz"
DSURQE_WM=f"{rabies_path}/DSURQE_40micron_eroded_WM_mask.nii.gz"
DSURQE_CSF=f"{rabies_path}/DSURQE_40micron_eroded_CSF_mask.nii.gz"
DSURQE_VASC=f"{rabies_path}/vascular_mask.nii.gz"
DSURQE_LABELS=f"{rabies_path}/DSURQE_40micron_labels.nii.gz"
DSURQE_MAPPINGS=f"{rabies_path}/DSURQE_40micron_R_mapping.csv"
DSURQE_ICA=f"{rabies_path}/melodic_IC.nii.gz"

EPICOMMON_ANAT=f"{rabies_path}/EPICOMMON/EPICOMMON_atlas/EPICOMMON_template.nii.gz"
EPICOMMON_MASK=f"{rabies_path}/EPICOMMON/EPICOMMON_atlas/EPICOMMON_brain_mask.nii.gz"
EPICOMMON_WM=f"{rabies_path}/EPICOMMON/EPICOMMON_atlas/EPICOMMON_WM_mask.nii.gz"
EPICOMMON_CSF=f"{rabies_path}/EPICOMMON/EPICOMMON_atlas/EPICOMMON_CSF_mask.nii.gz"
EPICOMMON_VASC=f"{rabies_path}/EPICOMMON/EPICOMMON_atlas/EPICOMMON_vascular_mask.nii.gz"
EPICOMMON_LABELS=f"{rabies_path}/EPICOMMON/EPICOMMON_atlas/EPICOMMON_labels.nii.gz"
EPICOMMON_ICA=f"{rabies_path}/EPICOMMON/EPICOMMON_atlas/melodic_IC_resampled.nii.gz"

RABIES_SEED_NAMES = [
        'ACA_seed', 'ECT_AI_seed', 'HY_seed', 'ORB_limbic_seed', 'SS_frontal_seed', 
        'AMYG_seed', 'RSP_seed', 'THAL_seed', 'basal_ganglia_seed', 'HIP_seed', 
        'MO_seed', 'SS_dorsal_seed', 'VIS_seed']

def execute_workflow(args=None, return_workflow=False):
    # generates the parser CLI and execute the workflow based on specified parameters.
    parser = get_parser()
    opts = read_parser(parser, args)

    # convert all input paths to absolute paths
    for arg in vars(opts):
        attr = getattr(opts, arg)
        if isinstance(attr,pathlib.Path):
            # convert to string since some downstream code needs cannot process pathlib.Path as input
            setattr(opts, arg, str(attr.resolve()))

    if not os.path.isdir(opts.output_dir):
        os.makedirs(opts.output_dir)

    log = prep_logging(opts, opts.output_dir)

    # verify default template installation
    install_DSURQE(log)

    from .__version__ import __version__
    log.info('Running RABIES - version: '+__version__)

    # print complete CLI command
    args = 'CLI INPUTS: \n'
    for arg in vars(opts):
        input = f'-> {arg} = {getattr(opts, arg)} \n'
        args += input
    log.info(args)

    # inclusion/exclusion list are incompatible parameters
    if (not opts.inclusion_ids[0]=='all') and (not opts.exclusion_ids[0]=='none'):
        raise ValueError(f"""
           Either an inclusion list (--inclusion_ids) or exclusion list (--exclusion_ids)
           can be provided, not both.
           """)
    
    ###ITK THREAD MANAGEMENT###
    if opts.num_ITK_threads=='optimal':
        num_ITK_threads=1 # until the number of scan is determined, this is set to 1 by default
        os.environ['RABIES_ITK_THREADS_STATE']='ON'
    elif opts.num_ITK_threads=='off':
        num_ITK_threads=1 # set to 1 to maximize nipype parallelization
        os.environ['RABIES_ITK_THREADS_STATE']='OFF' # this means run_command() won't set constraints on the number of threads
    else:
        num_ITK_threads=int(opts.num_ITK_threads)
        os.environ['RABIES_ITK_THREADS_STATE']='ON'
    # this is set as an os.environ variable instead of a nipype node input, to avoid re-running nodes when changing this parameter
    os.environ['RABIES_ITK_NUM_THREADS']=str(num_ITK_threads)

    if not opts.num_ITK_threads=='off':
        # python parallelization is entirely handled outside of SITK, so individual commands should inherit 1 CPU 
        sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(1)

    if str(opts.data_type) == 'int16':
        opts.data_type = sitk.sitkInt16
    elif str(opts.data_type) == 'int32':
        opts.data_type = sitk.sitkInt32
    elif str(opts.data_type) == 'float32':
        opts.data_type = sitk.sitkFloat32
    elif str(opts.data_type) == 'float64':
        opts.data_type = sitk.sitkFloat64
    else:
        raise ValueError('Invalid --data_type provided.')

    if str(opts.interpolation) == 'Linear':
        opts.interpolation_sitk = sitk.sitkLinear
    elif str(opts.interpolation) == 'BSpline3':
        opts.interpolation_sitk = sitk.sitkBSpline3
    elif str(opts.interpolation) == 'LanczosSinc':
        opts.interpolation_sitk = sitk.sitkLanczosWindowedSinc
    elif str(opts.interpolation) == 'CosineSinc':
        opts.interpolation_sitk = sitk.sitkCosineWindowedSinc
    else:
        raise ValueError('Invalid input for --interpolation.')

    if opts.rabies_stage == 'preprocess':
        workflow = preprocess(opts, log)
    elif opts.rabies_stage == 'confound_correction':
        workflow = confound_correction(opts, log)
    elif opts.rabies_stage == 'analysis':
        workflow = analysis(opts, log)
    else:
        parser.print_help()
    workflow.base_dir = opts.output_dir

    # the cli parameters are saved after workflow has been prepared, since they have to be modified during workflow preparation
    cli_file = f'{opts.output_dir}/rabies_{opts.rabies_stage}.pkl'
    with open(cli_file, 'wb') as handle:
        pickle.dump(opts, handle, protocol=pickle.HIGHEST_PROTOCOL)

    if return_workflow:
        return workflow
    try:
        log.info(f'Running workflow with {opts.plugin} plugin.')
        # execute workflow, with plugin_args limiting the cluster load for parallel execution
        graph_out = workflow.run(plugin=opts.plugin, plugin_args={'max_jobs': 50, 'dont_resubmit_completed_jobs': True,
                                                      'n_procs': opts.local_threads, 'qsub_args': f'-pe smp {str(opts.min_proc)}'})
        # save the workflow execution
        workflow_file = f'{opts.output_dir}/rabies_{opts.rabies_stage}_workflow.pkl'
        with open(workflow_file, 'wb') as handle:
            pickle.dump(graph_out, handle, protocol=pickle.HIGHEST_PROTOCOL)
        
    except Exception as e:
        log.critical(f'RABIES failed: {e}')
        raise


def prep_logging(opts, output_folder):
    cli_file = f'{output_folder}/rabies_{opts.rabies_stage}.pkl'
    if os.path.isfile(cli_file) and not opts.force:
        raise ValueError(f"""
            A previous run was indicated by the presence of {cli_file}.
            This can lead to inconsistencies between previous outputs and the log files.
            To prevent this, we recommend removing previous datasinks from the {opts.rabies_stage} 
            RABIES stage. To continue with your execution, the {cli_file} file must be  
            removed (use --force to automatically do so).
            """)

    # remove old versions of the log if already existing
    log_path = f'{output_folder}/rabies_{opts.rabies_stage}.log'
    if os.path.isfile(log_path):
        os.remove(log_path)

    config.update_config({'logging': {'log_directory': output_folder,
                                    'log_to_file': True}})

    # setting workflow logging level
    if opts.verbose==0:
        level="WARNING"
    elif opts.verbose==1:
        level="INFO"
    elif opts.verbose>=2:
        level="DEBUG"
        config.enable_debug_mode()
    else:
        raise ValueError(f"--verbose must be provided an integer of 0 or above. {opts.verbose} was provided instead.")

    # nipype has hard-coded 'nipype.log' filename; we rename it after it is created, and change the handlers
    logging.update_logging(config)
    os.rename(f'{output_folder}/pypeline.log', log_path)
    # change the handlers path to the desired file
    for logger in logging.loggers.keys():
        log = logging.getLogger(logger)
        handler = log.handlers[0]
        handler.baseFilename = log_path

    # set the defined level of verbose
    log = logging.getLogger('nipype.workflow')
    log.setLevel(level)
    log.debug('Debug ON')
    return log


def preprocess(opts, log):

    if not os.path.isdir(opts.bids_dir):
        raise ValueError("The provided BIDS data path doesn't exists.")
    else:
        # print the input data directory tree
        log.info("INPUT BIDS DATASET:  \n" + list_files(str(opts.bids_dir)))
    
    # if the default template is not used, then brain mask input is required, 
    # and other optional files are set to None if no input was provided 
    # to block downstream operations dependent on those inputs, but allow preprocessing nevertheless
    if not str(opts.anat_template)==DSURQE_ANAT:
        if str(opts.brain_mask)==DSURQE_MASK:
            raise ValueError("The default anatomical template was changed, but not the brain mask "
                             "- it is necessary to provide a new brain mask matching the template.")
        # make sure we have absolute paths
        opts.anat_template = os.path.abspath(opts.anat_template)
        opts.brain_mask = os.path.abspath(opts.brain_mask)
        
        for opt_key,default_file in zip(['WM_mask','CSF_mask','vascular_mask'],
                                         [DSURQE_WM,DSURQE_CSF,DSURQE_VASC]):
            
            opt_file = getattr(opts, opt_key)
            if str(opt_file)==default_file:
                opt_file=None
            else:
                opt_file = os.path.abspath(opt_file) # make sure we have absolute paths
            setattr(opts, opt_key, opt_file)
    else:
        opts.anat_template = os.path.abspath(opts.anat_template)
        opts.brain_mask = os.path.abspath(opts.brain_mask)

    # if --bold_only, the default atlas files change to EPI versions
    if opts.bold_only:
        for opt_key,default_file,EPI_file in zip(['anat_template','brain_mask','WM_mask','CSF_mask','vascular_mask'],
                                         [DSURQE_ANAT, DSURQE_MASK, DSURQE_WM,DSURQE_CSF,DSURQE_VASC],
                                         [EPICOMMON_ANAT, EPICOMMON_MASK, EPICOMMON_WM,EPICOMMON_CSF,EPICOMMON_VASC]):
            opt_file = getattr(opts, opt_key)
            if str(opt_file)==default_file:
                setattr(opts, opt_key, EPI_file)
                log.info(f'With --bold_only, default --{opt_key} changed to {EPI_file}')

    # final check of template file formats
    for opt_key,check_binary in zip(['anat_template', 'brain_mask', 'WM_mask','CSF_mask','vascular_mask'],
                                    [False,True,True,True,True]):
        opt_file = getattr(opts, opt_key)
        if opt_file is not None: # some masks might be set to None
            if not os.path.isfile(opt_file):
                raise ValueError(f"--{opt_key} file {opt_file} doesn't exists.")
            opt_file = convert_to_RAS(
                str(opt_file), opts.output_dir+'/template_files')

            if check_binary:
                check_binary_masks(opt_file)
            check_template_overlap(opts.anat_template, opt_file)
            setattr(opts, opt_key, opt_file)

    if not opts.inherit_unbiased_template=='none':
        opts.inherit_unbiased_template = os.path.abspath(opts.inherit_unbiased_template)
        if not os.path.isdir(opts.inherit_unbiased_template):
            raise ValueError(f"--inherit_unbiased_template path {opts.inherit_unbiased_template} doesn't exists.")

    check_resampling_syntax(opts.nativespace_resampling)
    check_resampling_syntax(opts.commonspace_resampling)
    check_resampling_syntax(opts.anatomical_resampling)

    # prepare nativespace/commonspace resampling arguments
    if opts.resampling_space=='both':
        opts.generate_commonspace = True
        opts.generate_nativespace = True
    elif opts.resampling_space=='common_only':
        opts.generate_commonspace = True
        opts.generate_nativespace = False
    elif opts.resampling_space=='native_only':
        opts.generate_commonspace = False
        opts.generate_nativespace = True

    # write boilerplate
    boilerplate_file = f'{opts.output_dir}/boilerplate.txt'

    methods,ref_string = preprocess_boilerplate(opts)
    txt_boilerplate="#######PREPROCESSING\n\n"+methods+ref_string+'\n\n'
    with open(boilerplate_file, "w") as text_file:
        text_file.write(txt_boilerplate)

    from rabies.preprocess_pkg.main_wf import init_main_wf
    workflow = init_main_wf(opts.bids_dir, opts.output_dir, opts)

    return workflow


def confound_correction(opts, log):

    if opts.conf_list:
        raise ValueError("--conf_list parameter is deprecated, and was replaced by --nuisance_regressors.")

    if opts.edge_cutoff == 0 and (opts.highpass is not None):
        log.warning(
            "\n############################################# WARNING\n"
            "Highpass filtering will be applied without removing timepoints at each edge "
            "of acquisition. This may introduce edge artefacts. We recommend removing "
            "~30sec at both end of the acquisition for a filter of 0.01Hz."
            "\n############################################# WARNING\n")

    cli_file = f'{opts.preprocess_out}/rabies_preprocess.pkl'
    with open(cli_file, 'rb') as handle:
        preprocess_opts = pickle.load(handle)

    boilerplate_file = f'{opts.output_dir}/boilerplate_confound_correction.txt'
    methods,ref_string = confound_correction_boilerplate(opts)
    txt_boilerplate="#######CONFOUND CORRECTION\n\n"+methods+ref_string+'\n\n'
    with open(boilerplate_file, "w") as text_file:
        text_file.write(txt_boilerplate)

    from rabies.confound_correction_pkg.main_wf import init_main_confound_correction_wf
    workflow = init_main_confound_correction_wf(preprocess_opts, opts)

    return workflow


def analysis(opts, log):

    cli_file = f'{opts.confound_correction_out}/rabies_confound_correction.pkl'
    with open(cli_file, 'rb') as handle:
        confound_correction_opts = pickle.load(handle)

    cli_file = f'{confound_correction_opts.preprocess_out}/rabies_preprocess.pkl'
    with open(cli_file, 'rb') as handle:
        preprocess_opts = pickle.load(handle)

    labels_file = opts.ROI_labels_file
    if str(labels_file)==DSURQE_LABELS:
        if str(preprocess_opts.anat_template)==DSURQE_ANAT:
            pass
        elif str(preprocess_opts.anat_template)==EPICOMMON_ANAT:
            file=EPICOMMON_LABELS
            opts.ROI_labels_file=EPICOMMON_LABELS
            log.info('With --bold_only, default --ROI_labels_file changed to '+file)
        else:
            opts.ROI_labels_file = None # set to None so that no computation is attempted using the labels
    else:
        if not os.path.isfile(labels_file):
            raise ValueError(f"--ROI_labels_file file {labels_file} doesn't exists.")
        # need to convert to RAS first, since the template file was also converted
        labels_file = convert_to_RAS(
            str(labels_file), preprocess_opts.output_dir+'/template_files')
        check_template_overlap(preprocess_opts.anat_template, labels_file)
        opts.ROI_labels_file = labels_file

    if preprocess_opts.bold_only and str(opts.prior_maps)==DSURQE_ICA:
        file=EPICOMMON_ICA
        opts.prior_maps=file
        log.info('With --bold_only, default --prior_maps changed to '+file)
        
    seed_file_list = []
    seed_name_list = []
    for seed in opts.seed_list:
        if seed in RABIES_SEED_NAMES:
            if preprocess_opts.bold_only:
                seed_file = f'{rabies_path}/DSURQE_seeds/EPICOMMON_resampled/{seed}_left_EPICOMMON_resampled.nii.gz'
            else:
                seed_file = f'{rabies_path}/DSURQE_seeds/{seed}_left.nii.gz'
            seed_name = seed
        else:
            seed_file = pathlib.Path(seed).resolve() # convert to absolute path
            seed_name = seed_file.name.rsplit(".nii")[0]
        if not os.path.isfile(seed_file):
            raise ValueError(
                f"Provide seed file path {seed_file} doesn't exists.")

        check_template_overlap(preprocess_opts.anat_template, seed_file)

        seed_file_list.append(seed_file)
        seed_name_list.append(seed_name)
    opts.seed_list = seed_file_list # override old input so that pre-built seeds are now files
    opts.seed_name_list = seed_name_list # store for developing iterables later

    from rabies.analysis_pkg.main_wf import init_main_analysis_wf
    workflow = init_main_analysis_wf(preprocess_opts, confound_correction_opts, opts)

    return workflow

def install_DSURQE(log):

    install = False
    # verifies whether default template files are installed and installs them otherwise
    for f in [DSURQE_ANAT,DSURQE_MASK,DSURQE_WM,DSURQE_CSF,DSURQE_VASC,DSURQE_LABELS,DSURQE_MAPPINGS,DSURQE_ICA,
              EPICOMMON_ANAT,EPICOMMON_MASK,EPICOMMON_WM,EPICOMMON_CSF,EPICOMMON_VASC,EPICOMMON_LABELS,EPICOMMON_ICA]:
        if not os.path.isfile(f):
            install = True
    for seed in RABIES_SEED_NAMES:
        seed_file = f'{rabies_path}/DSURQE_seeds/EPICOMMON_resampled/{seed}_left_EPICOMMON_resampled.nii.gz'
        if not os.path.isfile(seed_file):
            install = True
        seed_file = f'{rabies_path}/DSURQE_seeds/{seed}_left.nii.gz'
        if not os.path.isfile(seed_file):
            install = True

    if install:
        from rabies.utils import run_command
        log.info(
            "SOME FILES FROM THE DEFAULT TEMPLATE ARE MISSING. THEY WILL BE INSTALLED BEFORE FURTHER PROCESSING.")
        rc,c_out = run_command(f'install_DSURQE.sh {rabies_path}', verbose=True)


def check_binary_masks(mask):
    img = sitk.ReadImage(mask)
    array = sitk.GetArrayFromImage(img)
    if ((array != 1)*(array != 0)).sum() > 0:
        raise ValueError(
            f"The file {mask} is not a binary mask. Non-binary masks cannot be processed.")


def check_template_overlap(template, mask):
    template_img = sitk.ReadImage(template)
    mask_img = sitk.ReadImage(mask)
    if not template_img.GetOrigin() == mask_img.GetOrigin() and template_img.GetDirection() == mask_img.GetDirection():
        raise ValueError(
            f"The file {mask} does not appear to overlap with provided template {template}.")


def check_resampling_syntax(resampling):
    if resampling == 'inputs_defined':
        return
    try:
        if not 'x' in resampling:
            raise
        shape = resampling.split('x')
        if not len(shape) == 3:
            raise
        spacing = (float(shape[0]), float(shape[1]), float(shape[2]))
    except:
        raise ValueError(
            f"Resampling {resampling} must follow the format 'dim1xdim2xdim3', e.g. '0.1x0.1x0.1', (in mm) following the RAS axis convention.")


def list_files(startpath):
    string = ''
    for root, dirs, files in os.walk(startpath):
        level = root.replace(startpath, '').count(os.sep)
        indent = ' ' * 4 * (level)
        string += f'{indent}{os.path.basename(root)}/ \n'
        subindent = ' ' * 4 * (level + 1)
        for f in files:
            string += f'{subindent}{f} \n'
    return string



import os
import pathlib
from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu
from nipype import Function

from rabies.utils import ResampleVolumes
from .analysis_functions import run_group_ICA, run_DR_ICA, run_FC_matrix, seed_based_FC


def init_analysis_wf(opts, name="analysis_wf"):

    workflow = pe.Workflow(name=name)
    subject_inputnode = pe.Node(niu.IdentityInterface(
        fields=['CR_dict_file', 'maps_dict_file', 'token', 'native_to_commonspace_transform_list', 'native_to_commonspace_inverse_list']), name='subject_inputnode')
    group_inputnode = pe.Node(niu.IdentityInterface(
        fields=['bold_file_list', 'commonspace_mask', 'commonspace_template', 'token']), name='group_inputnode')
    outputnode = pe.Node(niu.IdentityInterface(fields=['group_ICA_dir', 'IC_file', 'dual_regression_timecourse_csv',
                                                       'DR_nii_file', 'DR_nii_file_resampled', 'matrix_data_file', 'matrix_fig', 'corr_map_file', 'corr_map_file_resampled', 'seed_timecourse_csv', 'joined_corr_map_file', 'joined_seed_timecourse_csv',
                                                       'sub_token', 'group_token','NPR_prior_timecourse_csv', 'NPR_extra_timecourse_csv',
                                                       'NPR_prior_filename', 'NPR_extra_filename', 'NPR_optimize_report']), name='outputnode')

    # connect the nodes so that they exist even without running analysis
    workflow.connect([
        (subject_inputnode, outputnode, [
            ("token", "sub_token"),
            ]),
        (group_inputnode, outputnode, [
            ("token", "group_token"),
            ]),
        ])

    if len(opts.seed_list) > 0:
        seed_based_FC_node = pe.Node(Function(input_names=['CR_dict_file', 'maps_dict_file', 'seed_name'],
                                              output_names=['corr_map_file', 'seed_timecourse_csv'],
                                              function=seed_based_FC),
                                     name='seed_based_FC', mem_gb=1*opts.scale_min_memory)
        seed_based_FC_node.iterables = ('seed_name', opts.seed_name_list)

        # create a joinnode to provide also combined seed maps
        seed_FC_joinnode = pe.JoinNode(niu.IdentityInterface(fields=['joined_corr_map_file', 'joined_seed_timecourse_csv']),
                                                name='seed_FC_joinnode',
                                                joinsource='seed_based_FC',
                                                joinfield=['joined_corr_map_file', 'joined_seed_timecourse_csv'])


        workflow.connect([
            (subject_inputnode, seed_based_FC_node, [
                ("CR_dict_file", "CR_dict_file"),
                ("maps_dict_file", "maps_dict_file"),
                ]),
            (seed_based_FC_node, outputnode, [
                ("corr_map_file", "corr_map_file"),
                ("seed_timecourse_csv", "seed_timecourse_csv"),
                ]),
            (seed_based_FC_node, seed_FC_joinnode, [
                ("seed_timecourse_csv", "joined_seed_timecourse_csv"),
                ]),
            (seed_FC_joinnode, outputnode, [
                ("joined_corr_map_file", "joined_corr_map_file"),
                ("joined_seed_timecourse_csv", "joined_seed_timecourse_csv"),
                ]),
            ])


        if opts.resample_to_commonspace:
            SBC_transform_node = pe.Node(ResampleVolumes(
                resampling_dim='ref_file', interpolation=opts.interpolation_sitk,
                rabies_data_type=opts.data_type, apply_motcorr=False, clip_negative=False), 
                name='SBC_to_commonspace')
            
            workflow.connect([
                (subject_inputnode, SBC_transform_node, [
                    ("native_to_commonspace_transform_list", "transforms_3d_files"),
                    ("native_to_commonspace_inverse_list", "inverses_3d"),
                    ]),
                (group_inputnode, SBC_transform_node, [
                    ("commonspace_template", "ref_file"),
                    ]),
                (seed_based_FC_node, SBC_transform_node, [
                    ("corr_map_file", "in_file"),
                    ("corr_map_file", "name_source"),
                    ]),
                (SBC_transform_node, outputnode, [
                    ("resampled_file", "corr_map_file_resampled"),
                    ]),
                (SBC_transform_node, seed_FC_joinnode, [
                    ("resampled_file", "joined_corr_map_file"),
                    ]),
                ])
        else:
            workflow.connect([
                (seed_based_FC_node, seed_FC_joinnode, [
                    ("corr_map_file", "joined_corr_map_file"),
                    ]),
                ])



    if opts.DR_ICA or opts.data_diagnosis:

        DR_ICA = pe.Node(Function(input_names=['CR_dict_file', 'maps_dict_file', 'network_weighting'],
                                  output_names=['DR_maps_filename', 'dual_regression_timecourse_csv'],
                                  function=run_DR_ICA),
                         name='DR_ICA', mem_gb=1*opts.scale_min_memory)
        DR_ICA.inputs.network_weighting = opts.network_weighting

        workflow.connect([
            (subject_inputnode, DR_ICA, [
                ("CR_dict_file", "CR_dict_file"),
                ("maps_dict_file", "maps_dict_file"),
                ]),
            (DR_ICA, outputnode, [
                ("dual_regression_timecourse_csv", "dual_regression_timecourse_csv"),
                ("DR_maps_filename", "DR_nii_file"),
                ]),
            ])

        if opts.resample_to_commonspace:
            DR_transform_node = pe.Node(ResampleVolumes(
                resampling_dim='ref_file', interpolation=opts.interpolation_sitk,
                rabies_data_type=opts.data_type, apply_motcorr=False, clip_negative=False), 
                n_procs=int(os.environ['RABIES_ITK_NUM_THREADS']),
                name='DR_to_commonspace')
            
            workflow.connect([
                (subject_inputnode, DR_transform_node, [
                    ("native_to_commonspace_transform_list", "transforms_3d_files"),
                    ("native_to_commonspace_inverse_list", "inverses_3d"),
                    ]),
                (group_inputnode, DR_transform_node, [
                    ("commonspace_template", "ref_file"),
                    ]),
                (DR_ICA, DR_transform_node, [
                    ("DR_maps_filename", "in_file"),
                    ("DR_maps_filename", "name_source"),
                    ]),
                (DR_transform_node, outputnode, [
                    ("resampled_file", "DR_nii_file_resampled"),
                    ]),
                ])

    if opts.group_ica['apply']:
        group_ICA = pe.Node(Function(input_names=['bold_file_list', 'mask_file', 'dim', 'random_seed', 'background_image', 'disableMigp'],
                                     output_names=['out_dir', 'IC_file'],
                                     function=run_group_ICA),
                            name='group_ICA', mem_gb=1*opts.scale_min_memory)
        group_ICA.inputs.dim = opts.group_ica['dim']
        group_ICA.inputs.random_seed = opts.group_ica['random_seed']
        group_ICA.inputs.disableMigp = opts.group_ica['disableMigp']

        workflow.connect([
            (group_inputnode, group_ICA, [
                ("bold_file_list", "bold_file_list"),
                ("commonspace_mask", "mask_file"),
                ("commonspace_template", "background_image"),
                ]),
            (group_ICA, outputnode, [
                ("IC_file", "IC_file"),
                ("out_dir", "group_ICA_dir"),
                ]),
            ])

    if (opts.NPR_temporal_comp>-1) or (opts.NPR_spatial_comp>-1) or opts.optimize_NPR['apply']:
        from .analysis_functions import NeuralPriorRecovery

        NPR_node = pe.Node(NeuralPriorRecovery(
                            NPR_temporal_comp=opts.NPR_temporal_comp, 
                            NPR_spatial_comp=opts.NPR_spatial_comp, 
                            optimize_NPR_dict=opts.optimize_NPR, 
                            prior_bold_idx = opts.prior_bold_idx,
                            network_weighting=opts.network_weighting,
                            figure_format=opts.figure_format),
                            name='NPR_node', mem_gb=1*opts.scale_min_memory)

        workflow.connect([
            (subject_inputnode, NPR_node, [
                ("CR_dict_file", "CR_dict_file"),
                ("maps_dict_file", "maps_dict_file"),
                ]),
            (NPR_node, outputnode, [
                ("NPR_prior_timecourse_csv", "NPR_prior_timecourse_csv"),
                ("NPR_extra_timecourse_csv", "NPR_extra_timecourse_csv"),
                ("NPR_prior_filename", "NPR_prior_filename"),
                ("NPR_extra_filename", "NPR_extra_filename"),
                ]),
            ])
        
    if opts.optimize_NPR['apply']:
        workflow.connect([
            (NPR_node, outputnode, [
                ("optimize_report", "NPR_optimize_report"),
                ]),
            ])


    if opts.FC_matrix:
        FC_matrix = pe.Node(Function(input_names=['CR_dict_file', 'maps_dict_file', 'figure_format', 'roi_type'],
                                     output_names=['data_file', 'figname'],
                                     function=run_FC_matrix),
                            name='FC_matrix', mem_gb=1*opts.scale_min_memory)
        FC_matrix.inputs.roi_type = opts.ROI_type
        FC_matrix.inputs.figure_format = opts.figure_format

        workflow.connect([
            (subject_inputnode, FC_matrix, [
                ("CR_dict_file", "CR_dict_file"),
                ("maps_dict_file", "maps_dict_file"),
                ]),
            (FC_matrix, outputnode, [
                ("data_file", "matrix_data_file"),
                ("figname", "matrix_fig"),
                ]),
            ])

    return workflow



import os
from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu
from nipype import Function

from rabies.analysis_pkg.diagnosis_pkg.interfaces import ScanDiagnosis, DatasetDiagnosis
from rabies.analysis_pkg.diagnosis_pkg.diagnosis_functions import temporal_external_formating, spatial_external_formating


def init_diagnosis_wf(analysis_opts, nativespace_analysis, preprocess_opts, split_name_list, name="diagnosis_wf"):

    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(
        fields=['CR_dict_file', 'common_maps_dict_file', 'sub_maps_dict_file', 'analysis_dict', 'native_to_commonspace_transform_list', 'native_to_commonspace_inverse_list']), name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(fields=['figure_temporal_diagnosis', 'figure_spatial_diagnosis', 
                                                       'analysis_QC', 'temporal_info_csv', 'spatial_VE_nii', 'temporal_std_nii', 'GS_corr_nii', 'GS_cov_nii',
                                                       'CR_prediction_std_nii']), name='outputnode')

    ScanDiagnosis_node = pe.Node(ScanDiagnosis(prior_bold_idx=analysis_opts.prior_bold_idx,
        prior_confound_idx=analysis_opts.prior_confound_idx,
            plot_seed_frequencies=analysis_opts.plot_seed_frequencies,
            figure_format=analysis_opts.figure_format, 
            nativespace_analysis=nativespace_analysis, 
            interpolation=analysis_opts.interpolation_sitk,
            brainmap_percent_threshold=analysis_opts.brainmap_percent_threshold,
            rabies_data_type=analysis_opts.data_type,
            ),
        name='ScanDiagnosis')

    temporal_external_formating_node = pe.Node(Function(input_names=['temporal_info'],
                                            output_names=[
                                                'temporal_info_csv'],
                                        function=temporal_external_formating),
                                name='temporal_external_formating')

    spatial_external_formating_node = pe.Node(Function(input_names=['spatial_info'],
                                            output_names=[
                                                'VE_filename', 'std_filename', 'predicted_std_filename', 
                                                'GS_corr_filename', 'GS_cov_filename'],
                                        function=spatial_external_formating),
                                name='spatial_external_formating')
    workflow.connect([
        (inputnode, ScanDiagnosis_node, [
            ("CR_dict_file", "CR_dict_file"),
            ("common_maps_dict_file", "common_maps_dict_file"),
            ("sub_maps_dict_file", "sub_maps_dict_file"),
            ("analysis_dict", "analysis_dict"),
            ("native_to_commonspace_transform_list", "native_to_common_transforms"),
            ("native_to_commonspace_inverse_list", "native_to_common_inverses"),
            ]),
        (ScanDiagnosis_node, temporal_external_formating_node, [
            ("temporal_info", "temporal_info"),
            ]),
        (ScanDiagnosis_node, spatial_external_formating_node, [
            ("spatial_info", "spatial_info"),
            ]),
        ])

    workflow.connect([
        (ScanDiagnosis_node, outputnode, [
            ("figure_temporal_diagnosis", "figure_temporal_diagnosis"),
            ("figure_spatial_diagnosis", "figure_spatial_diagnosis"),
            ]),
        (temporal_external_formating_node, outputnode, [
            ("temporal_info_csv", "temporal_info_csv"),
            ]),
        (spatial_external_formating_node, outputnode, [
            ("VE_filename", "spatial_VE_nii"),
            ("std_filename", "temporal_std_nii"),
            ("predicted_std_filename", "CR_prediction_std_nii"),
            ("GS_corr_filename", "GS_corr_nii"),
            ("GS_cov_filename", "GS_cov_nii"),
            ]),
        ])

    if not len(split_name_list)<3:

        # this function prepares a dictionary with necessary scan-level inputs for group diagnosis
        def prep_scan_data(CR_dict_file, maps_dict_file, spatial_info, temporal_info):
            import pickle
            with open(CR_dict_file, 'rb') as handle:
                CR_data_dict = pickle.load(handle)
            with open(maps_dict_file, 'rb') as handle:
                maps_data_dict = pickle.load(handle)

            scan_data={}

            dict_keys = ['temporal_std', 'VE_spatial', 'predicted_std', 'GS_corr', 'GS_cov',
                            'DR_BOLD', 'NPR_maps', 'prior_maps', 'seed_map_list']
            for key in dict_keys:
                scan_data[key] = spatial_info[key]

            # prepare the network and confound timecourses
            scan_data['DR_confound_time'] = temporal_info['DR_confound']
            scan_data['DR_network_time'] = temporal_info['DR_bold']
            scan_data['NPR_network_time'] = temporal_info['NPR_time']
            scan_data['SBC_network_time'] = temporal_info['SBC_time']

            scan_data['FD_trace'] = CR_data_dict['CR_data_dict']['FD_trace']
            scan_data['FD_DVARS_corr'] = CR_data_dict['CR_data_dict']['FD_DVARS_corr']
            scan_data['tDOF'] = CR_data_dict['CR_data_dict']['tDOF']
            scan_data['CR_global_std'] = CR_data_dict['CR_data_dict']['CR_global_std']
            scan_data['VE_total_ratio'] = CR_data_dict['CR_data_dict']['VE_total_ratio']
            scan_data['voxelwise_mean'] = CR_data_dict['CR_data_dict']['voxelwise_mean']

            scan_data['name_source'] = CR_data_dict['name_source']
            scan_data['anat_ref_file'] = maps_data_dict['anat_ref_file']
            scan_data['mask_file'] = maps_data_dict['mask_file']
            return scan_data

        prep_scan_data_node = pe.Node(Function(input_names=['CR_dict_file', 'maps_dict_file', 'spatial_info', 'temporal_info'],
                                            output_names=['scan_data'],
                                        function=prep_scan_data),
                                name='prep_scan_data_node')

        # calculate the number of scans combined in diagnosis
        num_scan = len(split_name_list)
        num_procs = min(analysis_opts.local_threads, num_scan)

        data_diagnosis_split_joinnode = pe.JoinNode(niu.IdentityInterface(fields=['scan_data_list']),
                                                name='diagnosis_split_joinnode',
                                                joinsource='main_split',
                                                joinfield=['scan_data_list'],
                                                n_procs=num_procs, mem_gb=1*num_scan*analysis_opts.scale_min_memory)

        DatasetDiagnosis_node = pe.Node(DatasetDiagnosis(),
            name='DatasetDiagnosis',
            n_procs=num_procs, mem_gb=0.2*num_scan*analysis_opts.scale_min_memory) # 0.2Gb, since no timeseries are inputted here
        DatasetDiagnosis_node.inputs.seed_prior_maps = analysis_opts.seed_prior_list
        DatasetDiagnosis_node.inputs.outlier_threshold = analysis_opts.outlier_threshold
        DatasetDiagnosis_node.inputs.network_weighting = analysis_opts.network_weighting
        DatasetDiagnosis_node.inputs.scan_QC_thresholds = analysis_opts.scan_QC_thresholds
        DatasetDiagnosis_node.inputs.figure_format = analysis_opts.figure_format
        DatasetDiagnosis_node.inputs.extended_QC = analysis_opts.extended_QC
        DatasetDiagnosis_node.inputs.group_avg_prior = analysis_opts.group_avg_prior
        DatasetDiagnosis_node.inputs.brainmap_percent_threshold = analysis_opts.brainmap_percent_threshold
        DatasetDiagnosis_node.inputs.add_smoothing = analysis_opts.qc_smooth_maps

        workflow.connect([
            (inputnode, prep_scan_data_node, [
                ("CR_dict_file", "CR_dict_file"),
                ("common_maps_dict_file", "maps_dict_file"),
                ]),
            (ScanDiagnosis_node, prep_scan_data_node, [
                ("spatial_info", "spatial_info"),
                ("temporal_info", "temporal_info"),
                ]),
            (prep_scan_data_node, data_diagnosis_split_joinnode, [
                ("scan_data", "scan_data_list"),
                ]),
            (data_diagnosis_split_joinnode, DatasetDiagnosis_node, [
                ("scan_data_list", "scan_data_list"),
                ]),
            (DatasetDiagnosis_node, outputnode, [
                ("analysis_QC", "analysis_QC"),
                ]),
            ])
    else:
        from nipype import logging
        log = logging.getLogger('nipype.workflow')
        log.warning(
            "Cannot run statistics on a sample size smaller than 3, so dataset diagnosis is not run.")

    return workflow



from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu
from nipype import Function
import os

def init_cross_modal_reg_wf(opts, name='cross_modal_reg_wf'):
    # cross_modal_reg_head_start
    """
    The input volumetric EPI image is registered non-linearly to an associated structural MRI image.
    The non-linear transform estimates the correction for EPI susceptibility distortions (Wang et al., 2017).

    References:
        Wang, S., Peterson, D. J., Gatenby, J. C., Li, W., Grabowski, T. J., & Madhyastha, T. M. (2017). 
            Evaluation of Field Map and Nonlinear Registration Methods for Correction of Susceptibility Artifacts 
            in Diffusion MRI. Frontiers in Neuroinformatics, 11, 17.

    Command line interface parameters:                  
        --bold2anat_coreg BOLD2ANAT_COREG
                            Specify the registration script for cross-modal alignment between the EPI and structural
                            images. This operation is responsible for correcting EPI susceptibility distortions.
                            * masking: With this option, the brain masks obtained from the EPI inhomogeneity correction 
                            step are used to support registration.
                            *** Specify 'true' or 'false'. 
                            * brain_extraction: conducts brain extraction prior to registration using the EPI masks from 
                            inhomogeneity correction. This will enhance brain edge-matching, but requires good quality 
                            masks. This should be selected along the 'masking' option.
                            *** Specify 'true' or 'false'. 
                            * winsorize_lower_bound: the lower bound for the antsRegistration winsorize-image-intensities option, useful for fUS images with intensity outliers.
                            *** Specify a value between 0 and 1.
                            * winsorize_upper_bound: the upper bound for the antsRegistration winsorize-image-intensities option, useful for fUS images with intensity outliers.
                            *** Specify a value between 0 and 1.
                            * keep_mask_after_extract: If using brain_extraction, use the mask to compute the registration metric
                            within the mask only. Choose to prevent stretching of the images beyond the limit of the brain mask
                            (e.g. if the moving and target images don't have the same brain coverage).
                            *** Specify 'true' or 'false'.
                            * registration: Specify a registration script.
                            *** Rigid: conducts only rigid registration.
                            *** Affine: conducts Rigid then Affine registration.
                            *** SyN: conducts Rigid, Affine then non-linear registration.
                            *** no_reg: skip registration.
                            (default: masking=false,brain_extraction=false,winsorize_lower_bound=0.0,winsorize_upper_bound=1.0,keep_mask_after_extract=false,registration=SyN)

    Workflow:
        parameters
            opts: command line interface parameters

        inputs
            ref_bold_brain: volumetric EPI image to register
            anat_ref: the target structural image
            anat_mask: the brain mask of the structural image
            moving_mask: a EPI mask inherited from inhomogeneity correction

        outputs
            bold_to_anat_affine: affine transform from the EPI to the anatomical image
            bold_to_anat_warp: non-linear transform from the EPI to the anatomical image
            bold_to_anat_inverse_warp: inverse non-linear transform from the EPI to the anatomical image
            output_warped_bold: the EPI image warped onto the structural image
    """
    # cross_modal_reg_head_end

    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['ref_bold_brain', 'anat_ref', 'anat_mask', 'moving_mask']),
        name='inputnode'
    )

    outputnode = pe.Node(
        niu.IdentityInterface(fields=[
            'bold_to_anat_affine', 'bold_to_anat_warp', 'bold_to_anat_inverse_warp', 'output_warped_bold']),
        name='outputnode'
    )

    run_reg = pe.Node(Function(input_names=["reg_method", "brain_extraction", "keep_mask_after_extract", "moving_image", "moving_mask", "fixed_image",
                                            "fixed_mask", "winsorize_lower_bound", "winsorize_upper_bound", "rabies_data_type"],
                               output_names=['bold_to_anat_affine', 'bold_to_anat_warp',
                                             'bold_to_anat_inverse_warp', 'output_warped_bold'],
                               function=run_antsRegistration), name='EPI_Coregistration', mem_gb=3*opts.scale_min_memory,
                               n_procs=int(os.environ['RABIES_ITK_NUM_THREADS']))

    # don't use brain extraction without a moving mask
    run_reg.inputs.winsorize_lower_bound = opts.bold2anat_coreg['winsorize_lower_bound']
    run_reg.inputs.winsorize_upper_bound = opts.bold2anat_coreg['winsorize_upper_bound']
    run_reg.inputs.reg_method = opts.bold2anat_coreg['registration']
    run_reg.inputs.brain_extraction = opts.bold2anat_coreg['brain_extraction']
    run_reg.inputs.keep_mask_after_extract = opts.bold2anat_coreg['keep_mask_after_extract']
    run_reg.inputs.rabies_data_type = opts.data_type
    run_reg.plugin_args = {
        'qsub_args': f'-pe smp {str(3*opts.min_proc)}', 'overwrite': True}

    if opts.bold2anat_coreg['masking']:
        workflow.connect([
            (inputnode, run_reg, [
                ('moving_mask', 'moving_mask')]),
            ])


    workflow.connect([
        (inputnode, run_reg, [
            ('ref_bold_brain', 'moving_image'),
            ('anat_ref', 'fixed_image'),
            ('anat_mask', 'fixed_mask')]),
        (run_reg, outputnode, [
            ('bold_to_anat_affine', 'bold_to_anat_affine'),
            ('bold_to_anat_warp', 'bold_to_anat_warp'),
            ('bold_to_anat_inverse_warp', 'bold_to_anat_inverse_warp'),
            ('output_warped_bold', 'output_warped_bold'),
            ]),
        ])

    return workflow


def run_antsRegistration(reg_method, brain_extraction=False, keep_mask_after_extract=False, moving_image='NULL', moving_mask='NOMASK', fixed_image='NULL', fixed_mask='NOMASK', winsorize_lower_bound = 0.0, winsorize_upper_bound = 1.0, rabies_data_type=8):
    import os
    import pathlib  # Better path manipulation
    filename_split = pathlib.Path(moving_image).name.rsplit(".nii")

    from rabies.preprocess_pkg.registration import define_reg_script
    reg_call = define_reg_script(reg_method)

    if reg_method == 'Rigid' or reg_method == 'Affine' or reg_method == 'SyN':
        if not 'NULL' in fixed_mask:
            reg_call+=f" --fixed-mask {fixed_mask}"
        if not 'NULL' in moving_mask:
            reg_call+=f" --moving-mask {moving_mask}"
        if brain_extraction:
            reg_call+=" --mask-extract"
            if keep_mask_after_extract:
                reg_call+=" --keep-mask-after-extract"
        if winsorize_lower_bound>0.0 or winsorize_upper_bound<1.0: # only specify an input if they differ from the default
            reg_call+=f" --winsorize-image-intensities {winsorize_lower_bound},{winsorize_upper_bound}"

        command = f"{reg_call} --resampled-output {filename_split[0]}_output_warped_image.nii.gz {moving_image} {fixed_image} {filename_split[0]}_output_"
    else:
        command = f'{reg_call} {moving_image} {moving_mask} {fixed_image} {fixed_mask} {filename_split[0]}'
    from rabies.utils import run_command
    rc,c_out = run_command(command)

    cwd = os.getcwd()
    warped_image = f'{cwd}/{filename_split[0]}_output_warped_image.nii.gz'
    affine = f'{cwd}/{filename_split[0]}_output_0GenericAffine.mat'
    warp = f'{cwd}/{filename_split[0]}_output_1Warp.nii.gz'
    inverse_warp = f'{cwd}/{filename_split[0]}_output_1InverseWarp.nii.gz'
    if not os.path.isfile(warped_image) or not os.path.isfile(affine):
        raise ValueError(
            'REGISTRATION ERROR: OUTPUT FILES MISSING. Make sure the provided registration script runs properly.')
    if not os.path.isfile(warp) or not os.path.isfile(inverse_warp):
        from nipype import logging
        log = logging.getLogger('nipype.workflow')
        log.debug('No non-linear warp files as output. Assumes linear registration.')
        warp = 'NULL'
        inverse_warp = 'NULL'

    import SimpleITK as sitk
    sitk.WriteImage(sitk.ReadImage(warped_image, rabies_data_type), warped_image)

    return [affine, warp, inverse_warp, warped_image]


def define_reg_script(reg_option):
    import os
    if reg_option == 'Rigid':
        reg_call = "antsRegistration_affine_SyN.sh --linear-type rigid --skip-nonlinear"
    elif reg_option == 'Affine':
        reg_call = "antsRegistration_affine_SyN.sh --linear-type affine --skip-nonlinear"
    elif reg_option == 'SyN':
        reg_call = "antsRegistration_affine_SyN.sh --linear-type affine"
    elif reg_option == 'no_reg':
        reg_call = 'null_nonlin.sh'
    else:
        raise ValueError(
            'The registration option must be among Rigid,Affine,SyN or NULL.')
    return reg_call
